# This file is part of cloud-init. See LICENSE file for license information.

import copy
import re
from functools import partial
from typing import Optional
from unittest import mock

import pytest

from cloudinit import util
from cloudinit.config.cc_rsyslog import (
    apply_rsyslog_changes,
    handle,
    load_config,
    parse_remotes_line,
    remotes_to_rsyslog_cfg,
)
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import skipUnlessJsonSchema
from tests.unittests.util import get_cloud


class TestLoadConfig:
    BASECFG = {
        "config_filename": "20-cloud-config.conf",
        "config_dir": "/etc/rsyslog.d",
        "service_reload_command": "auto",
        "configs": [],
        "remotes": {},
        "check_exe": "rsyslogd",
        "packages": ["rsyslog"],
        "install_rsyslog": False,
    }

    BSDCFG = {
        "config_filename": "20-cloud-config.conf",
        "config_dir": "/usr/local/etc/rsyslog.d",
        "service_reload_command": "auto",
        "configs": [],
        "remotes": {},
        "check_exe": "rsyslogd",
        "packages": ["rsyslog"],
        "install_rsyslog": False,
    }

    def test_legacy_full(self):
        cloud = get_cloud(distro="ubuntu", metadata={})
        cfg = copy.deepcopy(self.BASECFG)

        found = load_config(
            {
                "rsyslog": ["*.* @192.168.1.1"],
                "rsyslog_dir": "mydir",
                "rsyslog_filename": "myfilename",
            },
            distro=cloud.distro,
        )
        cfg.update(
            {
                "configs": ["*.* @192.168.1.1"],
                "config_dir": "mydir",
                "config_filename": "myfilename",
                "service_reload_command": "auto",
            }
        )

        assert found == cfg

    def test_legacy_defaults(self):
        cloud = get_cloud(distro="ubuntu", metadata={})
        cfg = copy.deepcopy(self.BASECFG)

        found = load_config(
            {"rsyslog": ["*.* @192.168.1.1"]}, distro=cloud.distro
        )
        cfg.update({"configs": ["*.* @192.168.1.1"]})
        assert found == cfg

    def test_new_defaults(self):
        cloud = get_cloud(distro="ubuntu", metadata={})
        cfg = copy.deepcopy(self.BASECFG)

        assert load_config({}, distro=cloud.distro) == cfg

    def test_new_bsd_defaults(self):

        # patch for ifconfig -a
        with mock.patch(
            "cloudinit.distros.networking.subp.subp", return_values=("", None)
        ):
            cloud = get_cloud(distro="freebsd", metadata={})
        assert load_config({}, distro=cloud.distro) == self.BSDCFG

    def test_new_configs(self):
        cfg = copy.deepcopy(self.BASECFG)

        cfgs = ["*.* myhost", "*.* my2host"]
        cloud = get_cloud(distro="ubuntu", metadata={})
        cfg.update({"configs": cfgs})
        assert (
            load_config({"rsyslog": {"configs": cfgs}}, distro=cloud.distro)
            == cfg
        )


class TestApplyChanges:

    def test_simple(self, tmpdir):
        cfgline = "*.* foohost"
        changed = apply_rsyslog_changes(
            configs=[cfgline], def_fname="foo.cfg", cfg_dir=str(tmpdir)
        )

        fname = str(tmpdir.join("foo.cfg"))
        assert changed == [fname]
        assert util.load_text_file(fname) == cfgline + "\n"

    def test_multiple_files(self, tmpdir):
        configs = [
            "*.* foohost",
            {"content": "abc", "filename": "my.cfg"},
            {
                "content": "filefoo-content",
                "filename": str(tmpdir.join("mydir/mycfg")),
            },
        ]

        changed = apply_rsyslog_changes(
            configs=configs, def_fname="default.cfg", cfg_dir=str(tmpdir)
        )

        expected = [
            (tmpdir.join("default.cfg"), "*.* foohost\n"),
            (tmpdir.join("my.cfg"), "abc\n"),
            (tmpdir.join("mydir/mycfg"), "filefoo-content\n"),
        ]
        assert [str(f[0]) for f in expected] == changed
        actual = []
        for fname, _content in expected:
            util.load_text_file(fname)
            actual.append(
                (
                    fname,
                    util.load_text_file(str(fname)),
                )
            )
        assert expected == actual

    def test_repeat_def(self, tmpdir):
        configs = ["*.* foohost", "*.warn otherhost"]

        changed = apply_rsyslog_changes(
            configs=configs, def_fname="default.cfg", cfg_dir=str(tmpdir)
        )

        fname = tmpdir.join("default.cfg")
        assert changed == [str(fname)]

        expected_content = "\n".join([c for c in configs]) + "\n"
        found_content = util.load_text_file(str(fname))
        assert expected_content == found_content

    def test_multiline_content(self, tmpdir):
        configs = ["line1", "line2\nline3\n"]

        apply_rsyslog_changes(
            configs=configs, def_fname="default.cfg", cfg_dir=str(tmpdir)
        )

        fname = tmpdir.join("default.cfg")
        expected_content = "\n".join([c for c in configs])
        found_content = util.load_text_file(str(fname))
        assert expected_content == found_content


class TestParseRemotesLine:
    def test_valid_port(self):
        r = parse_remotes_line("foo:9")
        assert r.port == 9

    def test_invalid_port(self):
        with pytest.raises(ValueError):
            parse_remotes_line("*.* foo:abc")

    def test_valid_ipv6(self):
        r = parse_remotes_line("*.* [::1]")
        assert str(r) == "*.* @[::1]"

    def test_valid_ipv6_with_port(self):
        r = parse_remotes_line("*.* [::1]:100")
        assert r.port == 100
        assert r.addr == "::1"
        assert str(r) == "*.* @[::1]:100"

    def test_invalid_multiple_colon(self):
        with pytest.raises(ValueError):
            parse_remotes_line("*.* ::1:100")

    def test_name_in_string(self):
        r = parse_remotes_line("syslog.host", name="foobar")
        assert str(r) == "*.* @syslog.host # foobar"


class TestRemotesToSyslog:
    def test_simple(self):
        # str rendered line must appear in remotes_to_ryslog_cfg return
        mycfg = "*.* myhost"
        myline = str(parse_remotes_line(mycfg, name="myname"))
        r = remotes_to_rsyslog_cfg({"myname": mycfg})
        lines = r.splitlines()
        assert len(lines) == 1
        assert myline in lines

    def test_header_footer(self):
        header = "#foo head"
        footer = "#foo foot"
        r = remotes_to_rsyslog_cfg(
            {"myname": "*.* myhost"}, header=header, footer=footer
        )
        lines = r.splitlines()
        assert lines[0] == header
        assert lines[-1] == footer

    def test_with_empty_or_null(self):
        mycfg = "*.* myhost"
        myline = str(parse_remotes_line(mycfg, name="myname"))
        r = remotes_to_rsyslog_cfg(
            {"myname": mycfg, "removed": None, "removed2": ""}
        )
        lines = r.splitlines()
        assert len(lines) == 1
        assert myline in lines


class TestRsyslogSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            ({"rsyslog": {"remotes": {"any": "string"}}}, None),
            (
                {"rsyslog": {"unknown": "a"}},
                "Additional properties are not allowed",
            ),
            ({"rsyslog": {"configs": [{"filename": "a"}]}}, ""),
            (
                {
                    "rsyslog": {
                        "configs": [
                            {"filename": "a", "content": "a", "a": "a"}
                        ]
                    }
                },
                "",
            ),
            (
                {"rsyslog": {"remotes": ["a"]}},
                r"\['a'\] is not of type 'object'",
            ),
            ({"rsyslog": {"remotes": "a"}}, "'a' is not of type 'object"),
            ({"rsyslog": {"service_reload_command": "a"}}, ""),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(
                SchemaValidationError, match=error_msg if error_msg else None
            ):
                validate_cloudconfig_schema(config, get_schema(), strict=True)


class TestInvalidKeyType:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            (
                {"rsyslog": {"configs": 1}},
                (
                    "Invalid type for key `configs`. Expected type(s): "
                    "<class 'list'>. Current type: <class 'int'>"
                ),
            ),
            (
                {"rsyslog": {"configs": [], "config_dir": 1}},
                (
                    "Invalid type for key `config_dir`. Expected type(s): "
                    "<class 'str'>. Current type: <class 'int'>"
                ),
            ),
            (
                {"rsyslog": {"configs": [], "config_filename": True}},
                (
                    "Invalid type for key `config_filename`. Expected type(s):"
                    " <class 'str'>. Current type: <class 'bool'>"
                ),
            ),
            (
                {"rsyslog": {"service_reload_command": 3.14}},
                (
                    "Invalid type for key `service_reload_command`. "
                    "Expected type(s): (<class 'str'>, <class 'list'>). "
                    "Current type: <class 'float'>"
                ),
            ),
            (
                {"rsyslog": {"remotes": ["1", 2, 3.14]}},
                (
                    "Invalid type for key `remotes`. Expected type(s): "
                    "<class 'dict'>. Current type: <class 'list'>"
                ),
            ),
        ],
    )
    def test_invalid_key_types(self, config: dict, error_msg: Optional[str]):
        cloud = get_cloud(distro="ubuntu", metadata={})
        callable_ = partial(load_config, config, cloud.distro)
        if error_msg is None:
            callable_()
        else:
            with pytest.raises(ValueError, match=re.escape(error_msg)):
                callable_()


class TestInstallRsyslog:
    @mock.patch("cloudinit.config.cc_rsyslog.subp.which")
    def test_install_rsyslog_on_freebsd(self, m_which):
        config = {
            "install_rsyslog": True,
            "packages": ["rsyslog"],
            "check_exe": "rsyslogd",
        }
        # patch for ifconfig -a
        with mock.patch(
            "cloudinit.distros.networking.subp.subp", return_values=("", None)
        ):
            cloud = get_cloud(distro="freebsd", metadata={})
            m_which.return_value = None
            with mock.patch.object(
                cloud.distro, "install_packages"
            ) as m_install:
                handle("rsyslog", {"rsyslog": config}, cloud, [])
            m_which.assert_called_with(config["check_exe"])
            m_install.assert_called_with(config["packages"])

    @mock.patch("cloudinit.config.cc_rsyslog.util.is_BSD")
    @mock.patch("cloudinit.config.cc_rsyslog.subp.which")
    def test_no_install_rsyslog_with_check_exe(self, m_which, m_isbsd):
        config = {
            "install_rsyslog": True,
            "packages": ["rsyslog"],
            "check_exe": "rsyslogd",
        }
        cloud = get_cloud(distro="ubuntu", metadata={})
        m_isbsd.return_value = False
        m_which.return_value = "/usr/sbin/rsyslogd"
        with mock.patch.object(cloud.distro, "install_packages") as m_install:
            handle("rsyslog", {"rsyslog": config}, cloud, [])
        m_which.assert_called_with(config["check_exe"])
        m_install.assert_not_called()



# This file is part of cloud-init. See LICENSE file for license information.

import logging
from copy import deepcopy
from unittest import mock

import pytest

from cloudinit import cloud, helpers
from cloudinit.config import cc_resolv_conf
from cloudinit.config.cc_resolv_conf import generate_resolv_conf
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.helpers import cloud_init_project_dir
from tests.unittests.helpers import skipUnlessJsonSchema
from tests.unittests.util import MockDistro

LOG = logging.getLogger(__name__)
EXPECTED_HEADER = """\
# Your system has been configured with 'manage-resolv-conf' set to true.
# As a result, cloud-init has written this file with configuration data
# that it has been provided. Cloud-init, by default, will write this file
# a single time (PER_ONCE).
#\n\n"""


@pytest.mark.usefixtures("fake_filesystem")
class TestResolvConf:
    cfg = {"manage_resolv_conf": True, "resolv_conf": {}}

    def call_resolv_conf_handler(self, distro, conf, paths):
        ds = None
        cc = cloud.Cloud(ds, paths, {}, distro, None)
        cc_resolv_conf.handle("cc_resolv_conf", conf, cc, [])

    @mock.patch("cloudinit.config.cc_resolv_conf.templater.render_to_file")
    def test_resolv_conf_systemd_resolved(
        self, m_render_to_file, Distro, paths
    ):
        dist = Distro("photon", self.cfg)
        self.call_resolv_conf_handler(dist, self.cfg, paths)

        assert [
            mock.call(mock.ANY, "/etc/systemd/resolved.conf", mock.ANY)
        ] == m_render_to_file.call_args_list

    @mock.patch("cloudinit.config.cc_resolv_conf.templater.render_to_file")
    def test_resolv_conf_no_param(self, m_render_to_file, caplog, paths):
        tmp = deepcopy(self.cfg)
        tmp.pop("resolv_conf")
        self.call_resolv_conf_handler("photon", tmp, paths)

        assert (
            "manage_resolv_conf True but no parameters provided" in caplog.text
        )
        assert [
            mock.call(mock.ANY, "/etc/systemd/resolved.conf", mock.ANY)
        ] not in m_render_to_file.call_args_list

    @mock.patch("cloudinit.config.cc_resolv_conf.templater.render_to_file")
    def test_resolv_conf_manage_resolv_conf_false(
        self, m_render_to_file, caplog, Distro, paths
    ):
        tmp = deepcopy(self.cfg)
        tmp["manage_resolv_conf"] = False
        dist = Distro("photon", self.cfg)
        self.call_resolv_conf_handler(dist, tmp, paths)
        assert "'manage_resolv_conf' present but set to False" in caplog.text
        assert [
            mock.call(mock.ANY, "/etc/systemd/resolved.conf", mock.ANY)
        ] not in m_render_to_file.call_args_list

    @mock.patch("cloudinit.config.cc_resolv_conf.templater.render_to_file")
    def test_resolv_conf_etc_resolv_conf(
        self, m_render_to_file, Distro, paths
    ):
        dist = Distro("rhel", self.cfg)
        self.call_resolv_conf_handler(dist, self.cfg, paths)

        assert [
            mock.call(mock.ANY, "/etc/resolv.conf", mock.ANY)
        ] == m_render_to_file.call_args_list

    @mock.patch("cloudinit.config.cc_resolv_conf.templater.render_to_file")
    def test_resolv_conf_invalid_resolve_conf_fn(
        self, m_render_to_file, caplog, Distro, tmp_path
    ):
        ds = None
        dist = Distro("rhel", self.cfg)
        paths = helpers.Paths({"cloud_dir": str(tmp_path)})
        cc = cloud.Cloud(ds, paths, {}, dist, None)
        cc.distro.resolve_conf_fn = "bla"

        cc_resolv_conf.handle("rhel", self.cfg, cc, [])

        assert (
            "No template found, not rendering resolve configs" in caplog.text
        )

        assert [
            mock.call(mock.ANY, "/etc/resolv.conf", mock.ANY)
        ] not in m_render_to_file.call_args_list


class TestGenerateResolvConf:

    dist = MockDistro()
    tmpl_fn = cloud_init_project_dir("templates/resolv.conf.tmpl")

    @mock.patch("cloudinit.config.cc_resolv_conf.templater.render_to_file")
    def test_dist_resolv_conf_fn(self, m_render_to_file):
        self.dist.resolve_conf_fn = "/tmp/resolv-test.conf"
        generate_resolv_conf(
            self.tmpl_fn, mock.MagicMock(), self.dist.resolve_conf_fn
        )

        assert [
            mock.call(mock.ANY, self.dist.resolve_conf_fn, mock.ANY)
        ] == m_render_to_file.call_args_list

    @mock.patch("cloudinit.config.cc_resolv_conf.templater.render_to_file")
    def test_target_fname_is_used_if_passed(self, m_render_to_file):
        path = "/use/this/path"
        generate_resolv_conf(self.tmpl_fn, mock.MagicMock(), path)

        assert [
            mock.call(mock.ANY, path, mock.ANY)
        ] == m_render_to_file.call_args_list

    # Patch in templater so we can assert on the actual generated content
    @mock.patch("cloudinit.templater.util.write_file")
    # Parameterise with the value to be passed to generate_resolv_conf as the
    # params parameter, and the expected line after the header as
    # expected_extra_line.
    @pytest.mark.parametrize(
        "params,expected_extra_line",
        [
            # No options
            ({}, None),
            # Just a true flag
            ({"options": {"foo": True}}, "options foo"),
            # Just a false flag
            ({"options": {"foo": False}}, None),
            # Just an option
            ({"options": {"foo": "some_value"}}, "options foo:some_value"),
            # A true flag and an option
            (
                {"options": {"foo": "some_value", "bar": True}},
                "options bar foo:some_value",
            ),
            # Two options
            (
                {"options": {"foo": "some_value", "bar": "other_value"}},
                "options bar:other_value foo:some_value",
            ),
            # Everything
            (
                {
                    "options": {
                        "foo": "some_value",
                        "bar": "other_value",
                        "baz": False,
                        "spam": True,
                    }
                },
                "options spam bar:other_value foo:some_value",
            ),
        ],
    )
    def test_flags_and_options(
        self, m_write_file, params, expected_extra_line
    ):
        target_fn = "/etc/resolv.conf"
        generate_resolv_conf(self.tmpl_fn, params, target_fn)

        expected_content = EXPECTED_HEADER
        if expected_extra_line is not None:
            # If we have any extra lines, expect a trailing newline
            expected_content += "\n".join([expected_extra_line, ""])
        assert [
            mock.call(mock.ANY, expected_content, mode=mock.ANY)
        ] == m_write_file.call_args_list


class TestResolvConfSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # Valid
            ({"manage_resolv_conf": False}, None),
            ({"resolv_conf": {"options": {"any": "thing"}}}, None),
            # Invalid
            (
                {"manage_resolv_conf": "asdf"},
                "'asdf' is not of type 'boolean'",
            ),
            # What may be some common misunderstandings of the template
            (
                {"resolv_conf": {"nameserver": ["1.1.1.1"]}},
                "Additional properties are not allowed",
            ),
            (
                {"resolv_conf": {"nameservers": "1.1.1.1"}},
                "'1.1.1.1' is not of type 'array'",
            ),
            (
                {"resolv_conf": {"search": ["foo.com"]}},
                "Additional properties are not allowed",
            ),
            (
                {"resolv_conf": {"searchdomains": "foo.com"}},
                "'foo.com' is not of type 'array'",
            ),
            (
                {"resolv_conf": {"domain": ["foo.com"]}},
                r"\['foo.com'\] is not of type 'string'",
            ),
            (
                {"resolv_conf": {"sortlist": "1.2.3.4"}},
                "'1.2.3.4' is not of type 'array'",
            ),
            (
                {"resolv_conf": {"options": "timeout: 1"}},
                "'timeout: 1' is not of type 'object'",
            ),
            (
                {"resolv_conf": {"options": "rotate"}},
                "'rotate' is not of type 'object'",
            ),
            (
                {"resolv_conf": {"options": ["rotate"]}},
                r"\['rotate'\] is not of type 'object'",
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)



# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init
from socket import gaierror
from textwrap import dedent
from unittest.mock import patch

import pytest

from cloudinit import helpers
from cloudinit.net.dhcp import NoDHCPLeaseError
from cloudinit.sources import DataSourceHostname
from cloudinit.sources.DataSourceCloudStack import (
    CLOUD_STACK_DMI_NAME,
    DataSourceCloudStack,
    DataSourceCloudStackLocal,
    get_data_server,
    get_vr_address,
)
from tests.unittests.helpers import mock
from tests.unittests.util import MockDistro

SOURCES_PATH = "cloudinit.sources"
MOD_PATH = SOURCES_PATH + ".DataSourceCloudStack"
DS_PATH = MOD_PATH + ".DataSourceCloudStack"
DHCP_MOD_PATH = "cloudinit.net.dhcp"
FAKE_LEASE = {
    "interface": "eth0",
    "fixed-address": "192.168.0.1",
    "subnet-mask": "255.255.255.0",
    "routers": "192.168.0.1",
    "domain-name": "dhclient.local",
    "renew": "4 2017/07/27 18:02:30",
    "expire": "5 2017/07/28 07:08:15",
}

FAKE_LEASE_WITH_SERVER_IDENT = """\
lease {
  interface "eth0";
  fixed-address 10.0.0.5;
  server-name "DSM111070915004";
  option subnet-mask 255.255.255.0;
  option dhcp-lease-time 4294967295;
  option routers 10.0.0.1;
  option dhcp-message-type 5;
  option dhcp-server-identifier 168.63.129.16;
  option domain-name-servers 168.63.129.16;
  option dhcp-renewal-time 4294967295;
  option rfc3442-classless-static-routes 0,10,0,0,1,32,168,63,129,16,10,0,0,1,32,169,254,169,254,10,0,0,1;
  option unknown-245 a8:3f:81:10;
  option dhcp-rebinding-time 4294967295;
  renew 0 2160/02/17 02:22:33;
  rebind 0 2160/02/17 02:22:33;
  expire 0 2160/02/17 02:22:33;
}
"""  # noqa: E501


@pytest.fixture
def cloudstack_ds(request, paths):
    yield DataSourceCloudStack(sys_cfg={}, distro=MockDistro(), paths=paths)


@pytest.mark.usefixtures("dhclient_exists")
class TestCloudStackHostname:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, tmp_path):
        self.hostname = "vm-hostname"
        self.networkd_domainname = "networkd.local"
        self.isc_dhclient_domainname = "dhclient.local"

        get_hostname_parent = mock.MagicMock(
            return_value=DataSourceHostname(self.hostname, True)
        )
        mocker.patch(
            SOURCES_PATH + ".DataSource.get_hostname", get_hostname_parent
        )
        mocker.patch(
            DHCP_MOD_PATH + ".util.load_text_file",
            return_value=FAKE_LEASE_WITH_SERVER_IDENT,
        )
        # Mock cloudinit.net.dhcp.networkd_get_option_from_leases() method \
        # result since we don't have a DHCP client running
        networkd_get_option_from_leases = mock.MagicMock(
            return_value=self.networkd_domainname
        )
        mocker.patch(
            DHCP_MOD_PATH + ".networkd_get_option_from_leases",
            networkd_get_option_from_leases,
        )

    def test_get_domainname_networkd(self, cloudstack_ds):
        """
        Test if DataSourceCloudStack._get_domainname()
        gets domain name from systemd-networkd leases.
        """
        assert self.networkd_domainname == cloudstack_ds._get_domainname()

    def test_get_domainname_isc_dhclient(self, cloudstack_ds, mocker):
        """
        Test if DataSourceCloudStack._get_domainname()
        gets domain name from isc-dhcp-client leases
        """

        # Override systemd-networkd reply mock to None
        # to force the code to fallback to IscDhclient
        get_networkd_domain = mock.MagicMock(return_value=None)
        mocker.patch(
            DHCP_MOD_PATH + ".networkd_get_option_from_leases",
            get_networkd_domain,
        )

        with patch(
            MOD_PATH + ".util.load_text_file",
            return_value=dedent(
                """
                lease {
                  interface "eth0";
                  fixed-address 10.0.0.5;
                  server-name "DSM111070915004";
                  option subnet-mask 255.255.255.0;
                  option dhcp-lease-time 4294967295;
                  option routers 10.0.0.1;
                  option dhcp-message-type 5;
                  option dhcp-server-identifier 168.63.129.16;
                  option domain-name-servers 168.63.129.16;
                  option dhcp-renewal-time 4294967295;
                  option rfc3442-classless-static-routes """
                """0,10,0,0,1,32,168,63,129,16,10,0,0,1,32,169,254,"""
                """169,254,10,0,0,1;
                  option unknown-245 a8:3f:81:10;
                  option dhcp-rebinding-time 4294967295;
                """
                f"option domain-name {self.isc_dhclient_domainname};"
                """renew 0 2160/02/17 02:22:33;
                  rebind 0 2160/02/17 02:22:33;
                  expire 0 2160/02/17 02:22:33;
                }
                """
            ),
        ):
            result = cloudstack_ds._get_domainname()
        assert self.isc_dhclient_domainname == result

    def test_get_hostname_non_fqdn(self, cloudstack_ds):
        """
        Test get_hostname() method implementation
        with fqdn parameter=False.
        It should call the parent class method and should
        return its response intact.
        """
        expected = DataSourceHostname(self.hostname, True)
        result = cloudstack_ds.get_hostname(fqdn=False)
        assert expected == result

    def test_get_hostname_fqdn(self, cloudstack_ds):
        """
        Test get_hostname() method implementation
        with fqdn parameter=True.
        It should look for domain name in DHCP leases.
        """
        expected = DataSourceHostname(
            self.hostname + "." + self.networkd_domainname, True
        )
        result = cloudstack_ds.get_hostname(fqdn=True)
        assert expected == result

    def test_get_hostname_fqdn_fallback(self, cloudstack_ds, mocker):
        """
        Test get_hostname() when some error happens
        during domainname discovery.

        We mock both systemd-networkd discovery as None,
        And the IscDhclient not having domain-name option
        in the lease.

        It should return the hostname without domainname
        in such cases.
        """
        expected = DataSourceHostname(self.hostname, True)

        # Override systemd-networkd reply mock to None
        # to force the code to fallback to IscDhclient
        get_networkd_domain = mock.MagicMock(return_value=None)
        mocker.patch(
            DHCP_MOD_PATH + ".networkd_get_option_from_leases",
            get_networkd_domain,
        )

        mocker.patch(
            "cloudinit.distros.net.find_fallback_nic",
            return_value="eth0",
        )

        mocker.patch(
            MOD_PATH + ".dhcp.IscDhclient.get_newest_lease_file_from_distro",
            return_value=True,
        )

        mocker.patch(
            MOD_PATH + ".dhcp.IscDhclient.parse_leases", return_value=[]
        )

        lease = {
            "interface": "eth0",
            "fixed-address": "192.168.0.1",
            "subnet-mask": "255.255.255.0",
            "routers": "192.168.0.1",
            "renew": "4 2017/07/27 18:02:30",
            "expire": "5 2017/07/28 07:08:15",
        }
        mocker.patch(
            DHCP_MOD_PATH + ".IscDhclient.get_newest_lease",
            return_value=lease,
        )
        mocker.patch(
            DHCP_MOD_PATH + ".Dhcpcd.get_newest_lease", return_value=lease
        )

        cloudstack_ds.distro.fallback_interface = "eth0"
        with patch(MOD_PATH + ".util.load_text_file"):
            result = cloudstack_ds.get_hostname(fqdn=True)
            assert expected == result

    @pytest.mark.parametrize(
        "lease_key,expected_domain",
        [
            ("DOMAINNAME", "example.com"),
            ("Domain", "example.com"),
            ("domain-name", "example.com"),
        ],
    )
    def test__get_domainname_supports_all_casing_variants(
        self, lease_key, expected_domain
    ):
        """Ensure _get_domainname works with DOMAINNAME, Domain and
        domain-name."""
        # Mock the helper to return the domain only when the exact key is asked
        with patch(
            "cloudinit.net.dhcp.networkd_get_option_from_leases"
        ) as m_get:
            m_get.side_effect = lambda key, extra_keys=None: (
                "example.com   " if key == lease_key else None
            )

            ds = DataSourceCloudStack(
                {}, distro=MockDistro(), paths=helpers.Paths({})
            )
            assert ds._get_domainname() == expected_domain


class TestGetDataServer:
    @pytest.mark.parametrize(
        "addrinfo,expected,expected_log",
        (
            pytest.param(
                # Fake addrinfo
                [("_", "_", "_", "_", ("10.1.35.171", 80)), "_"],
                "10.1.35.171",
                None,
                id="success_on_dns_resolution",
            ),
            pytest.param(
                gaierror("Name or service not known"),
                None,
                "DNS Entry data-server not found",
                id="none_on_no_dns_resolution",
            ),
        ),
    )
    def test_data_server_from_dns(
        self, addrinfo, expected, expected_log, mocker, caplog
    ):
        """Lookup data-server from DNS."""
        if isinstance(addrinfo, Exception):
            mocker.patch(MOD_PATH + ".getaddrinfo", side_effect=addrinfo)
            assert expected == get_data_server()
        else:
            mocker.patch(MOD_PATH + ".getaddrinfo", return_value=addrinfo)
            assert expected is get_data_server()
        if expected_log:
            assert expected_log in caplog.text


@mock.patch(MOD_PATH + ".get_data_server", return_value="10.1.37.131")
@mock.patch(
    MOD_PATH + ".dhcp.networkd_get_option_from_leases",
    return_value="10.1.37.132",
)
class TestGetVrAddress:
    def test_get_vr_addr_from_dns(
        self, m_networkd_option_from_leases, m_get_data_server, caplog
    ):
        """cloud-init first obtains data-server if resolved by DNS"""
        assert "10.1.37.131" == get_vr_address(MockDistro())
        assert (
            "Found metadata server '10.1.37.131' via data-server DNS entry"
            in caplog.text
        )
        assert 0 == m_networkd_option_from_leases.call_count

    def test_get_vr_addr_from_networkd_leases(
        self, m_networkd_option_from_leases, m_get_data_server, mocker, caplog
    ):
        """When no DNS for data-server use networkd dhcp-server-identifier"""
        mocker.patch(MOD_PATH + ".get_data_server", return_value=None)
        assert "10.1.37.132" == get_vr_address(MockDistro())
        assert (
            "Found SERVER_ADDRESS '10.1.37.132' via networkd_leases"
            in caplog.text
        )
        m_networkd_option_from_leases.assert_called_once_with("SERVER_ADDRESS")


@pytest.mark.usefixtures("dhclient_exists")
@mock.patch(MOD_PATH + ".dmi.read_dmi_data", return_value=CLOUD_STACK_DMI_NAME)
class TestCloudStackPasswordFetching:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, tmp_path):
        mocker.patch(f"{MOD_PATH}.ec2")
        mocker.patch(f"{MOD_PATH}.uhelp")
        default_gw = "192.201.20.0"
        mocker.patch(
            DHCP_MOD_PATH + ".IscDhclient.get_newest_lease",
            return_value={
                "interface": "eth0",
                "fixed-address": "192.168.0.1",
                "subnet-mask": "255.255.255.0",
                "routers": "192.168.0.1",
                "renew": "4 2017/07/27 18:02:30",
                "expire": "5 2017/07/28 07:08:15",
                "dhcp-server-identifier": "168.63.129.16",
            },
        )
        get_newest_lease_file_from_distro = mock.MagicMock(return_value=None)
        mocker.patch(
            DHCP_MOD_PATH + ".IscDhclient.get_newest_lease",
            return_value={
                "interface": "eth0",
                "fixed-address": "192.168.0.1",
                "subnet-mask": "255.255.255.0",
                "routers": "192.168.0.1",
                "renew": "4 2017/07/27 18:02:30",
                "expire": "5 2017/07/28 07:08:15",
                "dhcp-server-identifier": "168.63.129.16",
            },
        )
        mocker.patch(
            DHCP_MOD_PATH + ".IscDhclient.get_newest_lease_file_from_distro",
            get_newest_lease_file_from_distro,
        )
        get_default_gw = mock.MagicMock(return_value=default_gw)
        mocker.patch(MOD_PATH + ".get_default_gateway", get_default_gw)

        get_networkd_server_address = mock.MagicMock(return_value=None)
        mocker.patch(
            MOD_PATH + ".dhcp.networkd_get_option_from_leases",
            get_networkd_server_address,
        )
        get_data_server = mock.MagicMock(return_value=None)
        mocker.patch(MOD_PATH + ".get_data_server", get_data_server)

    def _set_password_server_response(self, response_string, mocker):
        fake_resp = mock.MagicMock()
        fake_resp.contents = response_string.encode("utf-8")
        readurl_mock = mocker.patch(
            "cloudinit.sources.DataSourceCloudStack.uhelp.readurl",
            return_value=fake_resp,
        )
        return readurl_mock

    def test_empty_password_doesnt_create_config(
        self, _dmi, cloudstack_ds, mocker
    ):
        self._set_password_server_response("", mocker)
        cloudstack_ds.get_data()
        assert {} == cloudstack_ds.get_config_obj()

    def test_saved_password_doesnt_create_config(
        self, _dmi, cloudstack_ds, mocker
    ):
        self._set_password_server_response("saved_password", mocker)
        cloudstack_ds.get_data()
        assert {} == cloudstack_ds.get_config_obj()

    @mock.patch(DS_PATH + ".wait_for_metadata_service")
    def test_password_sets_password(self, m_wait, _dmi, cloudstack_ds, mocker):
        m_wait.return_value = True
        password = "SekritSquirrel"
        self._set_password_server_response(password, mocker)
        cloudstack_ds.get_data()
        assert password == cloudstack_ds.get_config_obj()["password"]

    @mock.patch(DS_PATH + ".wait_for_metadata_service")
    def test_bad_request_doesnt_stop_ds_from_working(
        self, m_wait, _dmi, cloudstack_ds, mocker
    ):
        m_wait.return_value = True
        self._set_password_server_response("bad_request", mocker)
        assert cloudstack_ds.get_data() is True

    def assertRequestTypesSent(self, readurl, expected_request_types):
        request_types = []
        for call in readurl.call_args_list:
            headers = call.kwargs.get("headers", {})
            domu_req = headers.get("DomU_Request")
            if domu_req:
                request_types.append(domu_req)
        assert expected_request_types == request_types

    @mock.patch(DS_PATH + ".wait_for_metadata_service")
    def test_valid_response_means_password_marked_as_saved(
        self, m_wait, _dmi, cloudstack_ds, mocker
    ):
        m_wait.return_value = True
        password = "SekritSquirrel"
        readurl = self._set_password_server_response(password, mocker)
        cloudstack_ds.get_data()
        self.assertRequestTypesSent(
            readurl, ["send_my_password", "saved_password"]
        )

    def _check_password_not_saved_for(
        self, response_string, cloudstack_ds, mocker
    ):
        readurl = self._set_password_server_response(
            response_string, mocker=mocker
        )
        with mock.patch(DS_PATH + ".wait_for_metadata_service") as m_wait:
            m_wait.return_value = True
            cloudstack_ds.get_data()
        self.assertRequestTypesSent(readurl, ["send_my_password"])

    def test_password_not_saved_if_empty(self, _dmi, cloudstack_ds, mocker):
        self._check_password_not_saved_for("", cloudstack_ds, mocker)

    def test_password_not_saved_if_already_saved(
        self, _dmi, cloudstack_ds, mocker
    ):
        self._check_password_not_saved_for(
            "saved_password", cloudstack_ds, mocker
        )

    def test_password_not_saved_if_bad_request(
        self, _dmi, cloudstack_ds, mocker
    ):
        self._check_password_not_saved_for(
            "bad_request", cloudstack_ds, mocker
        )


class TestDataSourceCloudStackLocal:

    @mock.patch(MOD_PATH + ".EphemeralIPNetwork", autospec=True)
    @mock.patch(MOD_PATH + ".net.find_fallback_nic")
    @mock.patch(MOD_PATH + ".get_vr_address", return_value="10.1.37.131")
    def test_local_datasource_fails_ephemeral_dhcp(
        self, m_get_vr_address, m_find_fallback_nic, m_dhcp, caplog, tmpdir
    ):
        distro = MockDistro()
        ds = DataSourceCloudStackLocal(
            {}, distro, helpers.Paths({"run_dir": tmpdir})
        )
        fallback_nic_cases = ["enp0s1", "enp0s1"]
        dhcp_results = [NoDHCPLeaseError, Exception("Something unexpected")]
        expected_logs = [
            (
                "Attempting DHCP on: enp0s1",
                "Unable to obtain a DHCP lease on enp0s1",
            ),
            (
                "Attempting DHCP on: enp0s1",
                "Failed fetching metadata service: Something unexpected",
            ),
        ]

        # Each of the above cases, except the first, increments the m_dhcp call
        # count by one. Therefore start at and expect 0, incrementing the
        # expectation with each iteration
        dhcp_module_call_count = 0
        for fallback_nic, dhcp_result, logs in zip(
            fallback_nic_cases, dhcp_results, expected_logs
        ):
            m_find_fallback_nic.return_value = fallback_nic
            m_dhcp.return_value.__enter__.side_effect = dhcp_result
            dhcp_module_call_count += 1
            assert m_dhcp.call_count == dhcp_module_call_count - 1
            assert ds._get_data() is False
            for msg in logs:
                assert msg in caplog.text

    @mock.patch(MOD_PATH + ".CloudStackPasswordServerClient.get_password")
    @mock.patch(SOURCES_PATH + ".helpers.ec2.get_instance_metadata")
    @mock.patch(SOURCES_PATH + ".helpers.ec2.get_instance_userdata")
    @mock.patch(DS_PATH + ".wait_for_metadata_service")
    @mock.patch(
        MOD_PATH + ".EphemeralIPNetwork",
        autospec=True,
    )
    @mock.patch(MOD_PATH + ".net.find_fallback_nic")
    # @mock.patch(MOD_PATH + ".get_vr_address", return_value="10.1.37.131")
    def test_local_datasource_success(
        self,
        # m_get_vr_address,
        m_find_fallback_nic,
        m_dhcp,
        m_wait_for_mds,
        m_get_userdata,
        m_get_metadata,
        m_get_password,
        tmpdir,
    ):
        distro = MockDistro()
        ds = DataSourceCloudStackLocal(
            {}, distro, helpers.Paths({"run_dir": tmpdir})
        )

        m_find_fallback_nic.return_value = "enp0s1"
        m_dhcp.return_value.__enter__.side_effect = (None,)
        m_wait_for_mds.return_value = (True,)
        m_get_userdata.return_value = "ud"
        m_get_metadata.return_value = "md"
        m_get_password.return_value = True

        assert ds._get_data() is True
        assert ds.userdata_raw == "ud"
        assert ds.metadata == "md"



# Copyright (C) 2009-2010 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Joe VLcek <JVLcek@RedHat.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"""
This test file exercises the code in sources DataSourceAltCloud.py
"""

import os
import shutil

import pytest

import cloudinit.sources.DataSourceAltCloud as dsac
from cloudinit import subp, util
from tests.unittests.helpers import mock

OS_UNAME_ORIG = getattr(os, "uname")


def _write_user_data_files(mount_dir, value):
    """
    Populate the deltacloud_user_data_file the user_data_file
    which would be populated with user data.
    """
    deltacloud_user_data_file = mount_dir + "/deltacloud-user-data.txt"
    user_data_file = mount_dir + "/user-data.txt"

    udfile = open(deltacloud_user_data_file, "w")
    udfile.write(value)
    udfile.close()
    os.chmod(deltacloud_user_data_file, 0o664)

    udfile = open(user_data_file, "w")
    udfile.write(value)
    udfile.close()
    os.chmod(user_data_file, 0o664)


def _remove_user_data_files(mount_dir, dc_file=True, non_dc_file=True):
    """
    Remove the test files: deltacloud_user_data_file and
    user_data_file
    """
    deltacloud_user_data_file = mount_dir + "/deltacloud-user-data.txt"
    user_data_file = mount_dir + "/user-data.txt"

    # Ignore any failures removing files that are already gone.
    if dc_file:
        try:
            os.remove(deltacloud_user_data_file)
        except OSError:
            pass

    if non_dc_file:
        try:
            os.remove(user_data_file)
        except OSError:
            pass


def _dmi_data(expected):
    """
    Spoof the data received over DMI
    """

    def _data(key):
        return expected

    return _data


@pytest.fixture
def force_x86_64():
    # We have a different code path for arm to deal with LP1243287
    # We have to switch arch to x86_64 to avoid test failure
    force_arch("x86_64")
    yield
    # Return back to original arch
    force_arch()


@pytest.mark.usefixtures("fake_filesystem", "force_x86_64")
class TestGetCloudType:
    """Test to exercise method: DataSourceAltCloud.get_cloud_type()"""

    def test_cloud_info_file_ioerror(self, caplog, paths, tmp_path):
        """Return UNKNOWN when /etc/sysconfig/cloud-info exists but errors."""
        assert "/etc/sysconfig/cloud-info" == dsac.CLOUD_INFO_FILE
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        # Attempting to read the directory generates IOError
        with mock.patch.object(dsac, "CLOUD_INFO_FILE", str(tmp_path)):
            assert "UNKNOWN" == dsrc.get_cloud_type()
        assert "[Errno 21] Is a directory: '%s'" % tmp_path in caplog.text

    def test_cloud_info_file(self, paths):
        """Return uppercase stripped content from /etc/sysconfig/cloud-info."""
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        cloud_info = "cloud-info"
        util.write_file(cloud_info, " OverRiDdeN CloudType ")
        # Attempting to read the directory generates IOError
        with mock.patch.object(dsac, "CLOUD_INFO_FILE", cloud_info):
            assert "OVERRIDDEN CLOUDTYPE" == dsrc.get_cloud_type()

    @mock.patch("cloudinit.dmi.read_dmi_data", side_effect=_dmi_data("RHEV"))
    def test_rhev(self, m_read_dmi_data, paths):
        """
        Test method get_cloud_type() for RHEVm systems.
        Forcing read_dmi_data return to match a RHEVm system: RHEV Hypervisor
        """
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert "RHEV" == dsrc.get_cloud_type()

    @mock.patch(
        "cloudinit.dmi.read_dmi_data",
        side_effect=_dmi_data("VMware Virtual Platform"),
    )
    def test_vsphere(self, m_read_dmi_data, paths):
        """
        Test method get_cloud_type() for vSphere systems.
        Forcing read_dmi_data return to match a vSphere system: RHEV Hypervisor
        """
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert "VSPHERE" == dsrc.get_cloud_type()

    @mock.patch(
        "cloudinit.dmi.read_dmi_data",
        side_effect=_dmi_data("Unrecognized Platform"),
    )
    def test_unknown(self, m_read_dmi_data, paths):
        """
        Test method get_cloud_type() for unknown systems.
        Forcing read_dmi_data return to match an unrecognized return.
        """
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert "UNKNOWN" == dsrc.get_cloud_type()


@pytest.mark.usefixtures("fake_filesystem")
class TestGetDataCloudInfoFile:
    """
    Test to exercise method: DataSourceAltCloud.get_data()
    With a contrived CLOUD_INFO_FILE
    """

    CLOUD_INFO_FILE = "cloud-info"

    def test_rhev(self, paths):
        """Success Test module get_data() forcing RHEV."""

        util.write_file(self.CLOUD_INFO_FILE, "RHEV")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        dsrc.user_data_rhevm = lambda: True
        with mock.patch.object(dsac, "CLOUD_INFO_FILE", self.CLOUD_INFO_FILE):
            assert True is dsrc.get_data()
        assert "altcloud" == dsrc.cloud_name
        assert "altcloud" == dsrc.platform_type
        assert "rhev (/dev/fd0)" == dsrc.subplatform

    def test_vsphere(self, paths):
        """Success Test module get_data() forcing VSPHERE."""

        util.write_file(self.CLOUD_INFO_FILE, "VSPHERE")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        dsrc.user_data_vsphere = lambda: True
        with mock.patch.object(dsac, "CLOUD_INFO_FILE", self.CLOUD_INFO_FILE):
            assert True is dsrc.get_data()
        assert "altcloud" == dsrc.cloud_name
        assert "altcloud" == dsrc.platform_type
        assert "vsphere (unknown)" == dsrc.subplatform

    def test_fail_rhev(self, paths):
        """Failure Test module get_data() forcing RHEV."""

        util.write_file(self.CLOUD_INFO_FILE, "RHEV")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        dsrc.user_data_rhevm = lambda: False
        with mock.patch.object(dsac, "CLOUD_INFO_FILE", self.CLOUD_INFO_FILE):
            assert False is dsrc.get_data()

    def test_fail_vsphere(self, paths):
        """Failure Test module get_data() forcing VSPHERE."""

        util.write_file(self.CLOUD_INFO_FILE, "VSPHERE")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        dsrc.user_data_vsphere = lambda: False
        with mock.patch.object(dsac, "CLOUD_INFO_FILE", self.CLOUD_INFO_FILE):
            assert False is dsrc.get_data()

    def test_unrecognized(self, paths):
        """Failure Test module get_data() forcing unrecognized."""

        util.write_file(self.CLOUD_INFO_FILE, "unrecognized")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        with mock.patch.object(dsac, "CLOUD_INFO_FILE", self.CLOUD_INFO_FILE):
            assert False is dsrc.get_data()


@pytest.fixture
def fake_dsca_cloud_info():
    dsac.CLOUD_INFO_FILE = "no such file"
    yield
    dsac.CLOUD_INFO_FILE = "/etc/sysconfig/cloud-info"


@pytest.mark.usefixtures(
    "fake_filesystem", "force_x86_64", "fake_dsca_cloud_info"
)
class TestGetDataNoCloudInfoFile:
    """
    Test to exercise method: DataSourceAltCloud.get_data()
    Without a CLOUD_INFO_FILE
    """

    @mock.patch(
        "cloudinit.dmi.read_dmi_data", side_effect=_dmi_data("RHEV Hypervisor")
    )
    def test_rhev_no_cloud_file(self, m_read_dmi_data, paths):
        """Test No cloud info file module get_data() forcing RHEV."""
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        dsrc.user_data_rhevm = lambda: True
        assert True is dsrc.get_data()

    @mock.patch(
        "cloudinit.dmi.read_dmi_data",
        side_effect=_dmi_data("VMware Virtual Platform"),
    )
    def test_vsphere_no_cloud_file(self, m_read_dmi_data, paths):
        """Test No cloud info file module get_data() forcing VSPHERE."""
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        dsrc.user_data_vsphere = lambda: True
        assert True is dsrc.get_data()

    @mock.patch(
        "cloudinit.dmi.read_dmi_data",
        side_effect=_dmi_data("Unrecognized Platform"),
    )
    def test_failure_no_cloud_file(self, m_read_dmi_data, paths):
        """Test No cloud info file module get_data() forcing unrecognized."""
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.get_data()


@pytest.fixture
def user_data(tmp_path):
    mount_dir = str(tmp_path)
    _write_user_data_files(mount_dir, "test user data")
    yield
    _remove_user_data_files(mount_dir)
    shutil.rmtree(mount_dir)


@pytest.mark.usefixtures("user_data")
@mock.patch(
    "cloudinit.sources.DataSourceAltCloud.modprobe_floppy",
    return_value=None,
)
@mock.patch(
    "cloudinit.sources.DataSourceAltCloud.util.udevadm_settle",
    return_value=("", ""),
)
@mock.patch("cloudinit.sources.DataSourceAltCloud.util.mount_cb")
class TestUserDataRhevm:
    """
    Test to exercise method: DataSourceAltCloud.user_data_rhevm()
    """

    def test_mount_cb_fails(
        self, m_mount_cb, m_udevadm_settle, m_modprobe_floppy, paths
    ):
        """Test user_data_rhevm() where mount_cb fails."""
        m_mount_cb.side_effect = util.MountFailedError("Failed Mount")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.user_data_rhevm()

    def test_modprobe_fails(
        self, m_mount_cb, m_udevadm_settle, m_modprobe_floppy, paths
    ):
        """Test user_data_rhevm() where modprobe fails."""
        m_modprobe_floppy.side_effect = subp.ProcessExecutionError(
            "Failed modprobe"
        )
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.user_data_rhevm()

    def test_no_modprobe_cmd(
        self, m_mount_cb, m_udevadm_settle, m_modprobe_floppy, paths
    ):
        """Test user_data_rhevm() with no modprobe command."""
        m_modprobe_floppy.side_effect = subp.ProcessExecutionError(
            "No such file or dir"
        )
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.user_data_rhevm()

    def test_udevadm_fails(
        self, m_mount_cb, m_udevadm_settle, m_modprobe_floppy, paths
    ):
        """Test user_data_rhevm() where udevadm fails."""
        m_udevadm_settle.side_effect = subp.ProcessExecutionError(
            "Failed settle."
        )
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.user_data_rhevm()

    def test_no_udevadm_cmd(
        self, m_mount_cb, m_udevadm_settle, m_modprobe_floppy, paths
    ):
        """Test user_data_rhevm() with no udevadm command."""
        m_udevadm_settle.side_effect = OSError("No such file or dir")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.user_data_rhevm()


@pytest.mark.usefixtures("user_data")
class TestUserDataVsphere:
    """
    Test to exercise method: DataSourceAltCloud.user_data_vsphere()
    """

    @mock.patch("cloudinit.sources.DataSourceAltCloud.util.find_devs_with")
    @mock.patch("cloudinit.sources.DataSourceAltCloud.util.mount_cb")
    def test_user_data_vsphere_no_cdrom(
        self, m_mount_cb, m_find_devs_with, paths
    ):
        """Test user_data_vsphere() where mount_cb fails."""

        m_mount_cb.return_value = []
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.user_data_vsphere()
        assert 0 == m_mount_cb.call_count

    @mock.patch("cloudinit.sources.DataSourceAltCloud.util.find_devs_with")
    @mock.patch("cloudinit.sources.DataSourceAltCloud.util.mount_cb")
    def test_user_data_vsphere_mcb_fail(
        self, m_mount_cb, m_find_devs_with, paths
    ):
        """Test user_data_vsphere() where mount_cb fails."""

        m_find_devs_with.return_value = ["/dev/mock/cdrom"]
        m_mount_cb.side_effect = util.MountFailedError("Unable To mount")
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        assert False is dsrc.user_data_vsphere()
        assert 1 == m_find_devs_with.call_count
        assert 1 == m_mount_cb.call_count

    @mock.patch("cloudinit.sources.DataSourceAltCloud.util.find_devs_with")
    @mock.patch("cloudinit.sources.DataSourceAltCloud.util.mount_cb")
    def test_user_data_vsphere_success(
        self, m_mount_cb, m_find_devs_with, tmp_path, paths
    ):
        """Test user_data_vsphere() where successful."""
        m_find_devs_with.return_value = ["/dev/mock/cdrom"]
        m_mount_cb.return_value = "raw userdata from cdrom"
        dsrc = dsac.DataSourceAltCloud({}, None, paths)
        cloud_info = tmp_path / "cloud-info"
        util.write_file(cloud_info, "VSPHERE")
        assert True is dsrc.user_data_vsphere()
        m_find_devs_with.assert_called_once_with("LABEL=CDROM")
        m_mount_cb.assert_called_once_with(
            "/dev/mock/cdrom", dsac.read_user_data_callback
        )
        with mock.patch.object(dsrc, "get_cloud_type", return_value="VSPHERE"):
            assert "vsphere (/dev/mock/cdrom)" == dsrc.subplatform


@pytest.mark.usefixtures("user_data")
class TestReadUserDataCallback:
    """
    Test to exercise method: DataSourceAltCloud.read_user_data_callback()
    """

    def test_callback_both(self, tmp_path):
        """Test read_user_data_callback() with both files."""
        assert "test user data" == dsac.read_user_data_callback(str(tmp_path))

    def test_callback_dc(self, tmp_path):
        """Test read_user_data_callback() with only DC file."""
        _remove_user_data_files(str(tmp_path), dc_file=False, non_dc_file=True)
        assert "test user data" == dsac.read_user_data_callback(str(tmp_path))

    def test_callback_non_dc(self, tmp_path):
        """Test read_user_data_callback() with only non-DC file."""
        _remove_user_data_files(str(tmp_path), dc_file=True, non_dc_file=False)
        assert "test user data" == dsac.read_user_data_callback(str(tmp_path))

    def test_callback_none(self, tmp_path):
        """Test read_user_data_callback() no files are found."""
        _remove_user_data_files(str(tmp_path))
        assert dsac.read_user_data_callback(str(tmp_path)) is None


def force_arch(arch=None):
    def _os_uname():
        return ("LINUX", "NODENAME", "RELEASE", "VERSION", arch)

    if arch:
        setattr(os, "uname", _os_uname)
    elif arch is None:
        setattr(os, "uname", OS_UNAME_ORIG)



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2012 Yahoo! Inc.
# Copyright (C) 2016 Amazon.com, Inc. or its affiliates.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
# Author: Andrew Jorgensen <ajorgens@amazon.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

# noqa: E402

import collections
import logging
import re
import sys
from typing import Any

from jinja2 import TemplateSyntaxError

from cloudinit import performance
from cloudinit import type_utils as tu
from cloudinit import util
from cloudinit.atomic_helper import write_file

# After bionic EOL, mypy==1.0.0 will be able to type-analyse dynamic
# base types, substitute this by:
# JUndefined: typing.Type
JUndefined: Any
try:
    from jinja2 import DebugUndefined as _DebugUndefined
    from jinja2 import Template as JTemplate

    JINJA_AVAILABLE = True
    JUndefined = _DebugUndefined
except (ImportError, AttributeError):
    JINJA_AVAILABLE = False
    JUndefined = object

LOG = logging.getLogger(__name__)
MISSING_JINJA_PREFIX = "CI_MISSING_JINJA_VAR/"


class JinjaSyntaxParsingException(TemplateSyntaxError):
    def __init__(
        self,
        error: TemplateSyntaxError,
    ) -> None:
        super().__init__(
            error.message or "unknown syntax error",
            error.lineno,
            error.name,
            error.filename,
        )
        self.source = error.source

    def __str__(self):
        """Avoid jinja2.TemplateSyntaxError multi-line __str__ format."""
        return self.format_error_message(
            syntax_error=self.message,
            line_number=self.lineno,
            line_content=self.source.splitlines()[self.lineno - 2].strip(),
        )

    @staticmethod
    def format_error_message(
        syntax_error: str,
        line_number: str,
        line_content: str = "",
    ) -> str:
        """Avoid jinja2.TemplateSyntaxError multi-line __str__ format."""
        line_content = f": {line_content}" if line_content else ""
        return JinjaSyntaxParsingException.message_template.format(
            syntax_error=syntax_error,
            line_number=line_number,
            line_content=line_content,
        )

    message_template = (
        "Unable to parse Jinja template due to syntax error: "
        "{syntax_error} on line {line_number}{line_content}"
    )


# Mypy, and the PEP 484 ecosystem in general, does not support creating
# classes with dynamic base types: https://stackoverflow.com/a/59636248
class UndefinedJinjaVariable(JUndefined):
    """Class used to represent any undefined jinja template variable."""

    def __str__(self):
        return "%s%s" % (MISSING_JINJA_PREFIX, self._undefined_name)

    def __sub__(self, other):
        other = str(other).replace(MISSING_JINJA_PREFIX, "")
        raise TypeError(
            'Undefined jinja variable: "{this}-{other}". Jinja tried'
            ' subtraction. Perhaps you meant "{this}_{other}"?'.format(
                this=self._undefined_name, other=other
            )
        )


@performance.timed("Rendering basic template")
def basic_render(content, params):
    """This does simple replacement of bash variable like templates.

    It identifies patterns like ${a} or $a and can also identify patterns like
    ${a.b} or $a.b which will look for a key 'b' in the dictionary rooted
    by key 'a'.
    """

    def replacer(match):
        # Only 1 of the 2 groups will actually have a valid entry.
        name = match.group(1)
        if name is None:
            name = match.group(2)
        if name is None:
            raise RuntimeError("Match encountered but no valid group present")
        path = collections.deque(name.split("."))
        selected_params = params
        while len(path) > 1:
            key = path.popleft()
            if not isinstance(selected_params, dict):
                raise TypeError(
                    "Can not traverse into"
                    " non-dictionary '%s' of type %s while"
                    " looking for subkey '%s'"
                    % (selected_params, tu.obj_name(selected_params), key)
                )
            selected_params = selected_params[key]
        key = path.popleft()
        if not isinstance(selected_params, dict):
            raise TypeError(
                "Can not extract key '%s' from non-dictionary '%s' of type %s"
                % (key, selected_params, tu.obj_name(selected_params))
            )
        return str(selected_params[key])

    return re.sub(
        r"\$\{([A-Za-z0-9_.]+)\}|\$([A-Za-z0-9_.]+)", replacer, content
    )


def detect_template(text):
    def jinja_render(content, params):
        # keep_trailing_newline is in jinja2 2.7+, not 2.6
        add = "\n" if content.endswith("\n") else ""
        try:
            with performance.Timed("Rendering jinja2 template"):
                return (
                    JTemplate(
                        content,
                        undefined=UndefinedJinjaVariable,
                        trim_blocks=True,
                        extensions=["jinja2.ext.do"],
                    ).render(**params)
                    + add
                )
        except TemplateSyntaxError as template_syntax_error:
            template_syntax_error.lineno += 1
            raise JinjaSyntaxParsingException(
                error=template_syntax_error,
            ) from template_syntax_error
        except Exception as unknown_error:
            raise unknown_error from unknown_error

    if text.find("\n") != -1:
        ident, rest = text.split("\n", 1)  # remove the first line
    else:
        ident = text
        rest = ""
    type_match = re.match(r"##\s*template:(.*)", ident, re.I)
    if not type_match:
        return ("basic", basic_render, text)
    else:
        template_type = type_match.group(1).lower().strip()
        if template_type not in ("jinja", "basic"):
            raise ValueError(
                "Unknown template rendering type '%s' requested"
                % template_type
            )
        if template_type == "jinja" and not JINJA_AVAILABLE:
            LOG.warning(
                "Jinja not available as the selected renderer for"
                " desired template, reverting to the basic renderer."
            )
            return ("basic", basic_render, rest)
        elif template_type == "jinja" and JINJA_AVAILABLE:
            return ("jinja", jinja_render, rest)
        # Only thing left over is the basic renderer (it is always available).
        return ("basic", basic_render, rest)


def render_from_file(fn, params):
    if not params:
        params = {}
    template_type, renderer, content = detect_template(util.load_text_file(fn))
    LOG.debug("Rendering content of '%s' using renderer %s", fn, template_type)
    return renderer(content, params)


def render_to_file(fn, outfn, params, mode=0o644):
    contents = render_from_file(fn, params)
    util.write_file(outfn, contents, mode=mode)


def render_string(content, params):
    """Render string"""
    if not params:
        params = {}
    _template_type, renderer, content = detect_template(content)
    return renderer(content, params)


def render_template(variant, template, output, is_yaml, prefix=None):
    contents = util.load_text_file(template)
    tpl_params = {"variant": variant, "prefix": prefix}
    contents = (render_string(contents, tpl_params)).rstrip() + "\n"
    if is_yaml:
        out = util.load_yaml(contents, default=True)
        if not out:
            raise RuntimeError(
                "Cannot render template file %s - invalid yaml." % template
            )
    if output == "-":
        sys.stdout.write(contents)
    else:
        write_file(output, contents, omode="w")



# This file is part of cloud-init. See LICENSE file for license information.

import textwrap
from unittest import mock

import yaml

from cloudinit.net import network_manager, network_state
from tests.unittests.helpers import dir2dict


def assert_equal_dict(expected_d, found_d):
    for p, c in expected_d.items():
        if p not in found_d:
            continue
        assert c == found_d[p]


class TestNetworkManagerRenderNetworkState:
    def _parse_network_state_from_config(self, config):
        with mock.patch("cloudinit.net.network_state.get_interfaces_by_mac"):
            config = yaml.safe_load(config)
            return network_state.parse_net_config_data(config)

    def test_bond_dns_baseline(self, tmpdir):

        config = textwrap.dedent(
            """\
            version: 1
            config:
              - mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: ens1f0np0
                subnets: []
                type: physical
              - mac_address: 'xx:xx:xx:xx:xx:01'
                mtu: 9000
                name: ens1f1np1
                subnets: []
                type: physical
              - bond_interfaces:
                  - ens1f0np0
                  - ens1f1np1
                mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: bond0
                params:
                  bond-ad_actor_sys_prio: 65535
                  bond-ad_actor_system: 00:00:00:00:00:00
                  bond-ad_select: bandwidth
                  bond-ad_user_port_key: 2
                  bond-all_slaves_active: 0
                  bond-arp_all_targets: any
                  bond-arp_interval: 100
                  bond-arp_ip_target: 127.0.0.1
                  bond-arp_validate: none
                  bond-downdelay: 0
                  bond-fail_over_mac: xx:xx:xx:xx:xx:01
                  bond-lacp_rate: slow
                  bond-miimon: 100
                  bond-min_links: 2
                  bond-mode: 802.3ad
                  bond-updelay: 0
                  bond-use_carrier: 1
                  bond-xmit_hash_policy: layer3+4
                subnets: []
                type: bond
              - name: bond0.123
                subnets:
                  - address: 0.0.0.0
                    ipv4: true
                    netmask: 255.255.255.0
                    prefix: 24
                    routes:
                      - gateway: 0.0.0.1
                        netmask: 0.0.0.0
                        network: 0.0.0.0
                    type: static
                type: vlan
                vlan_id: 123
                vlan_link: bond0
              - address: 1.1.1.1
                search: hostname1
                type: nameserver
            """
        )

        expected_config = {
            "/etc/NetworkManager/system-connections/cloud-init-ens1f0np0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f0np0
                uuid=99c4bf6c-1691-53c4-bfe8-abdcb90b278a
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:00

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-ens1f1np1.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f1np1
                uuid=2685ec2b-1c26-583d-a660-0ab24201fef3
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:01

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-bond0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init bond0
                uuid=54317911-f840-516b-a10d-82cb4c1f075c
                autoconnect-priority=120
                type=bond
                interface-name=bond0

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [bond]
                mode=802.3ad
                miimon=100
                xmit_hash_policy=layer3+4
                downdelay=0
                updelay=0
                fail_over_mac=xx:xx:xx:xx:xx:01
                ad_actor_sys_prio=65535
                ad_actor_system=00:00:00:00:00:00
                ad_select=bandwidth
                ad_user_port_key=2
                all_slaves_active=0
                arp_all_targets=any
                arp_interval=100
                arp_ip_target=127.0.0.1
                arp_validate=none
                lacp_rate=slow
                min_links=2
                use_carrier=1

                [ipv4]
                method=disabled
                may-fail=false

                [ipv6]
                method=disabled
                may-fail=false

                [ethernet]
                mtu=9000

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-bond0.123.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init bond0.123
                uuid=7541e7a5-450b-570b-b3e8-a7f9eced114a
                autoconnect-priority=120
                type=vlan
                interface-name=bond0.123

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [vlan]
                id=123
                parent=54317911-f840-516b-a10d-82cb4c1f075c

                [ipv4]
                method=manual
                may-fail=false
                address1=0.0.0.0/24
                route1=0.0.0.0/0,0.0.0.1
                dns=1.1.1.1;
                dns-search=hostname1;

                """
            ),
        }
        with mock.patch("cloudinit.net.get_interfaces_by_mac"):
            ns = self._parse_network_state_from_config(config)
            target = str(tmpdir)
            network_manager.Renderer().render_network_state(ns, target=target)
            rendered_content = dir2dict(target)
            assert_equal_dict(expected_config, rendered_content)

    def test_bond_dns_redacted_with_method_disabled(self, tmpdir):

        config = textwrap.dedent(
            """\
            version: 1
            config:
              - mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: ens1f0np0
                subnets: []
                type: physical
              - mac_address: 'xx:xx:xx:xx:xx:01'
                mtu: 9000
                name: ens1f1np1
                subnets: []
                type: physical
              - bond_interfaces:
                  - ens1f0np0
                  - ens1f1np1
                mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: bond0
                params:
                  bond-ad_actor_sys_prio: 65535
                  bond-ad_actor_system: 00:00:00:00:00:00
                  bond-ad_select: bandwidth
                  bond-ad_user_port_key: 2
                  bond-all_slaves_active: 0
                  bond-arp_all_targets: any
                  bond-arp_interval: 100
                  bond-arp_ip_target: 127.0.0.1
                  bond-arp_validate: none
                  bond-downdelay: 0
                  bond-fail_over_mac: xx:xx:xx:xx:xx:01
                  bond-lacp_rate: slow
                  bond-miimon: 100
                  bond-min_links: 2
                  bond-mode: 802.3ad
                  bond-updelay: 0
                  bond-use_carrier: 1
                  bond-xmit_hash_policy: layer3+4
                subnets: []
                type: bond
              - name: bond0.123
                subnets:
                  - address: 0.0.0.0
                    ipv4: true
                    netmask: 255.255.255.0
                    prefix: 24
                    routes:
                      - gateway: 0.0.0.1
                        netmask: 0.0.0.0
                        network: 0.0.0.0
                    type: ipv6_slaac  # !! to force ipvx.method to be disabled
                type: vlan
                vlan_id: 123
                vlan_link: bond0
              - address: 1.1.1.1
                search: hostname1
                type: nameserver
            """
        )

        expected_config = {
            "/etc/NetworkManager/system-connections/cloud-init-ens1f0np0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f0np0
                uuid=99c4bf6c-1691-53c4-bfe8-abdcb90b278a
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:00

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-ens1f1np1.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f1np1
                uuid=2685ec2b-1c26-583d-a660-0ab24201fef3
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:01

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-bond0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init bond0
                uuid=54317911-f840-516b-a10d-82cb4c1f075c
                autoconnect-priority=120
                type=bond
                interface-name=bond0

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [bond]
                mode=802.3ad
                miimon=100
                xmit_hash_policy=layer3+4
                downdelay=0
                updelay=0
                fail_over_mac=xx:xx:xx:xx:xx:01
                ad_actor_sys_prio=65535
                ad_actor_system=00:00:00:00:00:00
                ad_select=bandwidth
                ad_user_port_key=2
                all_slaves_active=0
                arp_all_targets=any
                arp_interval=100
                arp_ip_target=127.0.0.1
                arp_validate=none
                lacp_rate=slow
                min_links=2
                use_carrier=1

                [ipv4]
                method=disabled
                may-fail=false

                [ipv6]
                method=disabled
                may-fail=false

                [ethernet]
                mtu=9000

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-bond0.123.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init bond0.123
                uuid=7541e7a5-450b-570b-b3e8-a7f9eced114a
                autoconnect-priority=120
                type=vlan
                interface-name=bond0.123

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [vlan]
                id=123
                parent=54317911-f840-516b-a10d-82cb4c1f075c

                [ipv6]
                method=auto
                may-fail=false
                address1=0.0.0.0/24
                dns-search=hostname1;

                [ipv4]
                method=disabled
                route1=0.0.0.0/0,0.0.0.1

                """
            ),
        }
        with mock.patch("cloudinit.net.get_interfaces_by_mac"):
            ns = self._parse_network_state_from_config(config)
            target = str(tmpdir)
            network_manager.Renderer().render_network_state(ns, target=target)
            rendered_content = dir2dict(target)
            assert_equal_dict(expected_config, rendered_content)

    def test_bond_balance_tlb_baseline(self, tmpdir):

        config = textwrap.dedent(
            """\
            version: 1
            config:
              - mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: ens1f0np0
                subnets: []
                type: physical
              - mac_address: 'xx:xx:xx:xx:xx:01'
                mtu: 9000
                name: ens1f1np1
                subnets: []
                type: physical
              - bond_interfaces:
                  - ens1f0np0
                  - ens1f1np1
                mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: bond0
                params:
                  bond-active_slave: ens1f0np0
                  bond-all_slaves_active: 0
                  bond-arp_all_targets: any
                  bond-arp_interval: 100
                  bond-arp_ip_target: 127.0.0.1
                  bond-arp_validate: none
                  bond-downdelay: 0
                  bond-fail_over_mac: xx:xx:xx:xx:xx:01
                  bond-lp_interval: 1
                  bond-miimon: 100
                  bond-mode: balance-tlb
                  bond-primary: ens1f0np0
                  bond-primary_reselect: always
                  bond-resend_igmp: 1
                  bond-tlb_dynamic_lb: 1
                  bond-updelay: 0
                  bond-use_carrier: 1
                  bond-xmit_hash_policy: layer3+4
                subnets: []
                type: bond
            """
        )

        expected_config = {
            "/etc/NetworkManager/system-connections/cloud-init-ens1f0np0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f0np0
                uuid=99c4bf6c-1691-53c4-bfe8-abdcb90b278a
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:00

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-ens1f1np1.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f1np1
                uuid=2685ec2b-1c26-583d-a660-0ab24201fef3
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:01

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-bond0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init bond0
                uuid=54317911-f840-516b-a10d-82cb4c1f075c
                autoconnect-priority=120
                type=bond
                interface-name=bond0

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [bond]
                mode=balance-tlb
                miimon=100
                xmit_hash_policy=layer3+4
                downdelay=0
                updelay=0
                fail_over_mac=xx:xx:xx:xx:xx:01
                primary_reselect=always
                primary=ens1f0np0
                active_slave=ens1f0np0
                all_slaves_active=0
                arp_all_targets=any
                arp_interval=100
                arp_ip_target=127.0.0.1
                arp_validate=none
                lp_interval=1
                resend_igmp=1
                tlb_dynamic_lb=1
                use_carrier=1

                [ipv4]
                method=disabled
                may-fail=false

                [ipv6]
                method=disabled
                may-fail=false

                [ethernet]
                mtu=9000

                """
            ),
        }
        with mock.patch("cloudinit.net.get_interfaces_by_mac"):
            ns = self._parse_network_state_from_config(config)
            target = str(tmpdir)
            network_manager.Renderer().render_network_state(ns, target=target)
            rendered_content = dir2dict(target)
            assert_equal_dict(expected_config, rendered_content)

    def test_bond_balance_rr_baseline(self, tmpdir):

        config = textwrap.dedent(
            """\
            version: 1
            config:
              - mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: ens1f0np0
                subnets: []
                type: physical
              - mac_address: 'xx:xx:xx:xx:xx:01'
                mtu: 9000
                name: ens1f1np1
                subnets: []
                type: physical
              - bond_interfaces:
                  - ens1f0np0
                  - ens1f1np1
                mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: bond0
                params:
                  bond-all_slaves_active: 0
                  bond-arp_all_targets: any
                  bond-arp_interval: 100
                  bond-arp_ip_target: 127.0.0.1
                  bond-arp_validate: none
                  bond-downdelay: 0
                  bond-fail_over_mac: xx:xx:xx:xx:xx:01
                  bond-miimon: 100
                  bond-mode: balance-tlb
                  bond-packets_per_slave: 1
                  bond-resend_igmp: 1
                  bond-updelay: 0
                  bond-use_carrier: 1
                subnets: []
                type: bond
            """
        )

        expected_config = {
            "/etc/NetworkManager/system-connections/cloud-init-ens1f0np0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f0np0
                uuid=99c4bf6c-1691-53c4-bfe8-abdcb90b278a
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:00

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-ens1f1np1.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f1np1
                uuid=2685ec2b-1c26-583d-a660-0ab24201fef3
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:01

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-bond0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init bond0
                uuid=54317911-f840-516b-a10d-82cb4c1f075c
                autoconnect-priority=120
                type=bond
                interface-name=bond0

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [bond]
                mode=balance-tlb
                miimon=100
                downdelay=0
                updelay=0
                fail_over_mac=xx:xx:xx:xx:xx:01
                all_slaves_active=0
                arp_all_targets=any
                arp_interval=100
                arp_ip_target=127.0.0.1
                arp_validate=none
                packets_per_slave=1
                resend_igmp=1
                use_carrier=1

                [ipv4]
                method=disabled
                may-fail=false

                [ipv6]
                method=disabled
                may-fail=false

                [ethernet]
                mtu=9000

                """
            ),
        }
        with mock.patch("cloudinit.net.get_interfaces_by_mac"):
            ns = self._parse_network_state_from_config(config)
            target = str(tmpdir)
            network_manager.Renderer().render_network_state(ns, target=target)
            rendered_content = dir2dict(target)
            assert_equal_dict(expected_config, rendered_content)

    def test_bond_active_backup_baseline(self, tmpdir):

        config = textwrap.dedent(
            """\
            version: 1
            config:
              - mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: ens1f0np0
                subnets: []
                type: physical
              - mac_address: 'xx:xx:xx:xx:xx:01'
                mtu: 9000
                name: ens1f1np1
                subnets: []
                type: physical
              - bond_interfaces:
                  - ens1f0np0
                  - ens1f1np1
                mac_address: 'xx:xx:xx:xx:xx:00'
                mtu: 9000
                name: bond0
                params:
                  bond-active_slave: ens1f0np0
                  bond-all_slaves_active: 0
                  bond-arp_all_targets: any
                  bond-arp_interval: 100
                  bond-arp_ip_target: 127.0.0.1
                  bond-arp_validate: none
                  bond-downdelay: 0
                  bond-fail_over_mac: xx:xx:xx:xx:xx:01
                  bond-miimon: 100
                  bond-mode: balance-tlb
                  bond-num_grat_arp: 1
                  bond-num_unsol_na: 1
                  bond-resend_igmp: 1
                  bond-updelay: 0
                  bond-use_carrier: 1
                subnets: []
                type: bond
            """
        )

        expected_config = {
            "/etc/NetworkManager/system-connections/cloud-init-ens1f0np0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f0np0
                uuid=99c4bf6c-1691-53c4-bfe8-abdcb90b278a
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:00

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-ens1f1np1.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init ens1f1np1
                uuid=2685ec2b-1c26-583d-a660-0ab24201fef3
                autoconnect-priority=120
                type=ethernet
                slave-type=bond
                master=54317911-f840-516b-a10d-82cb4c1f075c

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [ethernet]
                mtu=9000
                mac-address=XX:XX:XX:XX:XX:01

                """
            ),
            "/etc/NetworkManager/system-connections/cloud-init-bond0.nmconnection": textwrap.dedent(  # noqa: E501
                """\
                # Generated by cloud-init. Changes will be lost.

                [connection]
                id=cloud-init bond0
                uuid=54317911-f840-516b-a10d-82cb4c1f075c
                autoconnect-priority=120
                type=bond
                interface-name=bond0

                [user]
                org.freedesktop.NetworkManager.origin=cloud-init

                [bond]
                mode=balance-tlb
                miimon=100
                num_grat_arp=1
                downdelay=0
                updelay=0
                fail_over_mac=xx:xx:xx:xx:xx:01
                active_slave=ens1f0np0
                all_slaves_active=0
                arp_all_targets=any
                arp_interval=100
                arp_ip_target=127.0.0.1
                arp_validate=none
                num_unsol_na=1
                resend_igmp=1
                use_carrier=1

                [ipv4]
                method=disabled
                may-fail=false

                [ipv6]
                method=disabled
                may-fail=false

                [ethernet]
                mtu=9000

                """
            ),
        }
        with mock.patch("cloudinit.net.get_interfaces_by_mac"):
            ns = self._parse_network_state_from_config(config)
            target = str(tmpdir)
            network_manager.Renderer().render_network_state(ns, target=target)
            rendered_content = dir2dict(target)
            assert_equal_dict(expected_config, rendered_content)



"""Tests for tools/render-template"""

import sys

import pytest

from cloudinit import subp, templater, util
from tests.helpers import cloud_init_project_dir
from tests.unittests.helpers import skipUnlessJinjaVersionGreaterThan

# TODO(Look to align with tools.render-template or cloudinit.distos.OSFAMILIES)
DISTRO_VARIANTS = [
    "amazon",
    "arch",
    "azurelinux",
    "centos",
    "debian",
    "eurolinux",
    "fedora",
    "freebsd",
    "gentoo",
    "mariner",
    "netbsd",
    "openbsd",
    "photon",
    "raspberry-pi-os",
    "rhel",
    "suse",
    "ubuntu",
    "unknown",
]


@pytest.mark.allow_subp_for(sys.executable)
class TestRenderCloudCfg:
    cmd = [sys.executable, cloud_init_project_dir("tools/render-template")]
    tmpl_path = cloud_init_project_dir("config/cloud.cfg.tmpl")
    init_path = cloud_init_project_dir("sysvinit/freebsd/dsidentify.tmpl")

    def test_variant_sets_distro_in_cloud_cfg_subp(self, tmpdir):
        outfile = tmpdir.join("outcfg").strpath

        subp.subp(self.cmd + ["--variant", "ubuntu", self.tmpl_path, outfile])
        with open(outfile) as stream:
            system_cfg = util.load_yaml(stream.read())
        assert system_cfg["system_info"]["distro"] == "ubuntu"

    def test_variant_sets_prefix_in_cloud_cfg_subp(self, tmpdir):
        outfile = tmpdir.join("outcfg").strpath

        subp.subp(
            self.cmd
            + [
                "--variant",
                "freebsd",
                "--prefix",
                "/usr/local",
                self.init_path,
                outfile,
            ]
        )
        with open(outfile) as stream:
            init_cfg = stream.readlines()
        assert 'command="/usr/local/lib/cloud-init/ds-identify"\n' in init_cfg

    @pytest.mark.parametrize("variant", (DISTRO_VARIANTS))
    def test_variant_sets_distro_in_cloud_cfg(self, variant, tmpdir):
        """Testing parametrized inputs with imported function saves ~0.5s per
        call versus calling as subp
        """
        outfile = tmpdir.join("outcfg").strpath

        templater.render_template(
            variant, self.tmpl_path, outfile, is_yaml=True
        )
        with open(outfile) as stream:
            system_cfg = util.load_yaml(stream.read())
        if variant == "unknown":
            variant = "ubuntu"  # Unknown is defaulted to ubuntu
        assert system_cfg["system_info"]["distro"] == variant

    @pytest.mark.parametrize("variant", (DISTRO_VARIANTS))
    def test_variant_sets_default_user_in_cloud_cfg(self, variant, tmpdir):
        """Testing parametrized inputs with imported function saves ~0.5s per
        call versus calling as subp
        """
        outfile = tmpdir.join("outcfg").strpath
        templater.render_template(
            variant, self.tmpl_path, outfile, is_yaml=True
        )
        with open(outfile) as stream:
            system_cfg = util.load_yaml(stream.read())

        default_user_exceptions = {
            "amazon": "ec2-user",
            "rhel": "cloud-user",
            "centos": "cloud-user",
            "raspberry-pi-os": "pi",
            "unknown": "ubuntu",
        }
        default_user = system_cfg["system_info"]["default_user"]["name"]
        assert default_user == default_user_exceptions.get(variant, variant)

    @pytest.mark.parametrize(
        "variant,renderers",
        (
            ("freebsd", ["freebsd"]),
            ("netbsd", ["netbsd"]),
            ("openbsd", ["openbsd"]),
            ("ubuntu", ["netplan", "eni", "sysconfig"]),
            ("raspberry-pi-os", ["netplan", "network-manager"]),
        ),
    )
    def test_variant_sets_network_renderer_priority_in_cloud_cfg(
        self, variant, renderers, tmpdir
    ):
        """Testing parametrized inputs with imported function saves ~0.5s per
        call versus calling as subp
        """
        outfile = tmpdir.join("outcfg").strpath
        templater.render_template(
            variant, self.tmpl_path, outfile, is_yaml=True
        )
        with open(outfile) as stream:
            system_cfg = util.load_yaml(stream.read())

        assert renderers == system_cfg["system_info"]["network"]["renderers"]


EXPECTED_DEBIAN = """\
deb testmirror testcodename main
deb-src testmirror testcodename main
deb testsecurity testcodename-security main
deb-src testsecurity testcodename-security main
deb testmirror testcodename-updates main
deb-src testmirror testcodename-updates main
deb testmirror testcodename-backports main
deb-src testmirror testcodename-backports main
"""

EXPECTED_DEBIAN_DEB822 = """\
Types: deb deb-src
URIs: testmirror
Suites: testcodename testcodename-updates testcodename-backports
Components: main
Signed-By: /usr/share/keyrings/debian-archive-keyring.gpg
Types: deb deb-src
URIs: testsecurity
Suites: testcodename-security
Components: main
Signed-By: /usr/share/keyrings/debian-archive-keyring.gpg
"""

EXPECTED_UBUNTU = """\
deb testmirror testcodename main restricted
deb testmirror testcodename-updates main restricted
deb testmirror testcodename universe
deb testmirror testcodename-updates universe
deb testmirror testcodename multiverse
deb testmirror testcodename-updates multiverse
deb testmirror testcodename-backports main restricted universe multiverse
deb testsecurity testcodename-security main restricted
deb testsecurity testcodename-security universe
deb testsecurity testcodename-security multiverse
"""

EXPECTED_UBUNTU_DEB822 = """\
Types: deb
URIs: testmirror
Suites: testcodename testcodename-updates testcodename-backports
Components: main universe restricted multiverse
Signed-By: /usr/share/keyrings/ubuntu-archive-keyring.gpg
Types: deb
URIs: testsecurity
Suites: testcodename-security
Components: main universe restricted multiverse
Signed-By: /usr/share/keyrings/ubuntu-archive-keyring.gpg
"""


class TestRenderSourcesList:
    @pytest.mark.parametrize(
        "template_path,expected",
        [
            pytest.param(
                "templates/sources.list.debian.tmpl",
                EXPECTED_DEBIAN,
                id="debian",
            ),
            pytest.param(
                "templates/sources.list.debian.deb822.tmpl",
                EXPECTED_DEBIAN_DEB822,
                id="debian_822",
            ),
            pytest.param(
                "templates/sources.list.ubuntu.tmpl",
                EXPECTED_UBUNTU,
                id="ubuntu",
            ),
            pytest.param(
                "templates/sources.list.ubuntu.deb822.tmpl",
                EXPECTED_UBUNTU_DEB822,
                id="ubuntu_822",
            ),
        ],
    )
    @skipUnlessJinjaVersionGreaterThan((3, 0, 0))
    def test_render_sources_list_templates(
        self, tmpdir, template_path, expected
    ):
        params = {
            "mirror": "testmirror",
            "security": "testsecurity",
            "codename": "testcodename",
        }
        template_path = cloud_init_project_dir(template_path)
        rendered = templater.render_string(open(template_path).read(), params)
        filtered = "\n".join(
            line
            for line in rendered.splitlines()
            if line.strip() and not line.strip().startswith("#")
        )
        assert filtered.strip() == expected.strip()



# This file is part of cloud-init. See LICENSE file for license information.

"""Tests for registering RHEL subscription via rh_subscription."""

import copy
import logging

import pytest

from cloudinit import subp
from cloudinit.config import cc_rh_subscription
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import (
    mock,
    skipUnlessJsonSchema,
    skipUnlessJsonSchemaVersionGreaterThan,
)

SUBMGR = cc_rh_subscription.SubscriptionManager
SUB_MAN_CLI = "cloudinit.config.cc_rh_subscription._sub_man_cli"
NAME = "cc_rh_subscription"


@mock.patch(SUB_MAN_CLI)
class TestHappyPath:

    CONFIG = {
        "rh_subscription": {
            "username": "scooby@do.com",
            "password": "scooby-snacks",
        }
    }

    CONFIG_FULL = {
        "rh_subscription": {
            "username": "scooby@do.com",
            "password": "scooby-snacks",
            "auto_attach": True,
            "service_level": "self-support",
            "add_pool": ["pool1", "pool2", "pool3"],
            "enable_repo": ["repo1", "repo2", "repo3"],
            "disable_repo": ["repo4", "repo5"],
            "release_version": "7.6b",
        }
    }

    def test_already_registered(self, m_sman_cli, caplog):
        """
        Emulates a system that is already registered. Ensure it gets
        a non-ProcessExecution error from is_registered()
        """
        cc_rh_subscription.handle(NAME, self.CONFIG, None, [])
        assert m_sman_cli.call_count == 1
        assert "System is already registered" in caplog.text

    @mock.patch.object(
        cc_rh_subscription.SubscriptionManager, "_set_release_version"
    )
    def test_simple_registration(
        self, m_set_release_version, m_sman_cli, caplog
    ):
        """
        Simple registration with username and password
        """
        reg = (
            "The system has been registered with ID:"
            " 12345678-abde-abcde-1234-1234567890abc"
        )
        m_sman_cli.side_effect = [subp.ProcessExecutionError, (reg, "bar")]
        cc_rh_subscription.handle(NAME, self.CONFIG, None, [])
        assert mock.call(["identity"]) in m_sman_cli.call_args_list
        assert (
            mock.call(
                [
                    "register",
                    "--username=scooby@do.com",
                    "--password=scooby-snacks",
                ],
                logstring_val=True,
            )
            in m_sman_cli.call_args_list
        )
        assert "rh_subscription plugin completed successfully" in caplog.text
        assert m_sman_cli.call_count == 2
        assert m_set_release_version.call_count == 0

    @pytest.mark.parametrize(
        "variable_name_separator",
        (
            pytest.param("_", id="update_repos_disable_with_none"),
            pytest.param(
                "-", id="same_functional_behavior_with_deprecated_keys"
            ),
        ),
    )
    @mock.patch.object(cc_rh_subscription.SubscriptionManager, "_getRepos")
    def test_update_repos_disable_with_none(
        self, m_get_repos, m_sman_cli, variable_name_separator
    ):
        cfg = copy.deepcopy(self.CONFIG)
        m_get_repos.return_value = ([], ["repo1"])

        enable_repo_key = "enable_repo".replace("_", variable_name_separator)
        disable_repo_key = "disable_repo".replace("_", variable_name_separator)
        cfg["rh_subscription"].update(
            {enable_repo_key: ["repo1"], disable_repo_key: None}
        )
        mysm = cc_rh_subscription.SubscriptionManager(cfg)
        assert True is mysm.update_repos()
        m_get_repos.assert_called_with()
        assert m_sman_cli.call_args_list == [
            mock.call(["repos", "--enable=repo1"])
        ]

    def test_full_registration(self, m_sman_cli, caplog):
        """
        Registration with auto_attach, service_level, adding pools,
        enabling and disabling yum repos and setting release_version
        """
        call_lists = []
        call_lists.append(["attach", "--pool=pool1", "--pool=pool3"])
        call_lists.append(
            ["repos", "--disable=repo5", "--enable=repo2", "--enable=repo3"]
        )
        call_lists.append(["attach", "--auto", "--servicelevel=self-support"])
        call_lists.append(["release", "--set=7.6b"])
        reg = (
            "The system has been registered with ID:"
            " 12345678-abde-abcde-1234-1234567890abc"
        )
        m_sman_cli.side_effect = [
            subp.ProcessExecutionError,
            (reg, "bar"),
            ("Service level set to: self-support", ""),
            ("pool1\npool3\n", ""),
            ("pool2\n", ""),
            ("", ""),
            ("Repo ID: repo1\nRepo ID: repo5\n", ""),
            ("Repo ID: repo2\nRepo ID: repo3\nRepo ID: repo4", ""),
            ("", ""),
            ("Release set to: 7.6b", ""),
        ]
        # to avoid deleting the actual cache files
        # (triggered by the presence of the release_version key)
        # on the host running the tests
        mock.patch("shutil.rmtree")

        cc_rh_subscription.handle(NAME, self.CONFIG_FULL, None, [])
        assert m_sman_cli.call_count == 10
        for call in call_lists:
            assert mock.call(call) in m_sman_cli.call_args_list
        assert "rh_subscription plugin completed successfully" in caplog.text


CONFIG_BADAUTOATTACH = {
    "rh_subscription": {
        "username": "scooby@do.com",
        "password": "scooby-snacks",
        "auto_attach": 3,
    }
}

CONFIG_SERVICE = {
    "rh_subscription": {
        "username": "scooby@do.com",
        "password": "scooby-snacks",
        "service_level": "self-support",
    }
}


@mock.patch(SUB_MAN_CLI)
class TestBadInput:
    SM = cc_rh_subscription.SubscriptionManager
    REG = (
        "The system has been registered with ID:"
        " 12345678-abde-abcde-1234-1234567890abc"
    )

    CONFIG_NO_PASSWORD = {"rh_subscription": {"username": "scooby@do.com"}}

    CONFIG_NO_KEY = {
        "rh_subscription": {
            "activation_key": "1234abcde",
        }
    }

    CONFIG_BADPOOL = {
        "rh_subscription": {
            "username": "scooby@do.com",
            "password": "scooby-snacks",
            "add_pool": "not_a_list",
        }
    }
    CONFIG_BADREPO = {
        "rh_subscription": {
            "username": "scooby@do.com",
            "password": "scooby-snacks",
            "enable_repo": "not_a_list",
        }
    }
    CONFIG_BAD_RELEASE_VERSION = {
        "rh_subscription": {
            "username": "scooby@do.com",
            "password": "scooby-snacks",
            "release_version": "bad_release_version",
        }
    }

    def assert_logged_warnings(self, warnings, caplog):
        missing = [
            w
            for w in warnings
            if (mock.ANY, logging.WARNING, w) not in caplog.record_tuples
        ]
        assert [] == missing, "Missing expected warnings."

    def test_no_password(self, m_sman_cli):
        """Attempt to register without the password key/value."""
        m_sman_cli.side_effect = [
            subp.ProcessExecutionError,
            (self.REG, "bar"),
        ]
        cc_rh_subscription.handle(NAME, self.CONFIG_NO_PASSWORD, None, [])
        assert m_sman_cli.call_count == 0

    def test_no_org(self, m_sman_cli, caplog):
        """Attempt to register without the org key/value."""
        m_sman_cli.side_effect = [subp.ProcessExecutionError]
        cc_rh_subscription.handle(NAME, self.CONFIG_NO_KEY, None, [])
        m_sman_cli.assert_called_with(["identity"])
        assert m_sman_cli.call_count == 1
        self.assert_logged_warnings(
            (
                "Unable to register system due to incomplete information.",
                "Use either activationkey and org *or* userid and password",
                "Registration failed or did not run completely",
                "rh_subscription plugin did not complete successfully",
            ),
            caplog,
        )

    @pytest.mark.parametrize(
        "auto_attach_cfg,warnings",
        (
            (
                CONFIG_SERVICE,
                [
                    "The service_level key must be used in conjunction with"
                    " the auto_attach key.  Please re-run with"
                    " auto_attach: True",
                    "rh_subscription plugin did not complete successfully",
                ],
            ),
            (
                CONFIG_BADAUTOATTACH,
                ["The key auto_attach must be a boolean value (True/False)"],
            ),
        ),
    )
    def test_service_level_without_auto(
        self, m_sman_cli, auto_attach_cfg, warnings, caplog
    ):
        """Attempt to register using service_level without auto_attach key."""
        m_sman_cli.side_effect = [
            subp.ProcessExecutionError,
            (self.REG, "bar"),
        ]
        cc_rh_subscription.handle(NAME, auto_attach_cfg, None, [])
        assert m_sman_cli.call_count == 1
        self.assert_logged_warnings(
            warnings,
            caplog,
        )

    def test_pool_not_a_list(self, m_sman_cli, caplog):
        """
        Register with pools that are not in the format of a list
        """
        m_sman_cli.side_effect = [
            subp.ProcessExecutionError,
            (self.REG, "bar"),
        ]
        cc_rh_subscription.handle(NAME, self.CONFIG_BADPOOL, None, [])
        assert m_sman_cli.call_count == 2
        self.assert_logged_warnings(
            (
                "Pools must in the format of a list",
                "rh_subscription plugin did not complete successfully",
            ),
            caplog,
        )

    def test_repo_not_a_list(self, m_sman_cli, caplog):
        """
        Register with repos that are not in the format of a list
        """
        m_sman_cli.side_effect = [
            subp.ProcessExecutionError,
            (self.REG, "bar"),
        ]
        cc_rh_subscription.handle(NAME, self.CONFIG_BADREPO, None, [])
        assert m_sman_cli.call_count == 2
        self.assert_logged_warnings(
            (
                "Repo IDs must in the format of a list.",
                "Unable to add or remove repos",
                "rh_subscription plugin did not complete successfully",
            ),
            caplog,
        )

    @mock.patch.object(
        cc_rh_subscription.SubscriptionManager, "_delete_packagemanager_cache"
    )
    def test_bad_release_version(self, m_delete_pm_cache, m_sman_cli, caplog):
        """
        Failure at setting release_version
        """
        m_sman_cli.side_effect = [
            subp.ProcessExecutionError,
            (self.REG, "bar"),
            subp.ProcessExecutionError,
        ]
        cc_rh_subscription.handle(
            NAME, self.CONFIG_BAD_RELEASE_VERSION, None, []
        )
        assert m_sman_cli.call_count == 3
        assert m_delete_pm_cache.call_count == 0
        expected_cmd = [
            "release",
            f"--set={self.CONFIG_BAD_RELEASE_VERSION['rh_subscription']['release_version']}",
        ]
        self.assert_logged_warnings(
            (
                f"Unable to set release_version using: {expected_cmd}",
                "rh_subscription plugin did not complete successfully",
            ),
            caplog,
        )

    @mock.patch("shutil.rmtree", side_effect=[PermissionError])
    def test_pm_cache_deletion_after_setting_release_version(
        self, m_rmtree, m_sman_cli, caplog
    ):
        """
        Failure at deleting package manager cache
        after setting release_version
        """
        good_release_ver_cfg = copy.deepcopy(self.CONFIG_BAD_RELEASE_VERSION)
        good_release_ver_cfg["rh_subscription"][
            "release_version"
        ] = "1.2Server"
        m_sman_cli.side_effect = [
            subp.ProcessExecutionError,
            (self.REG, "bar"),
            ("Release set to: 1.2Server", ""),
        ]
        cc_rh_subscription.handle(NAME, good_release_ver_cfg, None, [])
        # assert "rh_subscription plugin completed successfully" in caplog.text
        assert m_sman_cli.call_count == 3
        assert m_rmtree.call_args_list == [mock.call("/var/cache/dnf")]
        self.assert_logged_warnings(
            (
                "Unable to delete the package manager cache",
                "rh_subscription plugin did not complete successfully",
            ),
            caplog,
        )


class TestRhSubscriptionSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            (
                {"rh_subscription": {"bad": "input"}},
                "Additional properties are not allowed",
            ),
            (
                {"rh_subscription": {"add-pool": [1]}},
                "1 is not of type 'string'",
            ),
            (
                {"rh_subscription": {"add_pool": [1]}},
                "1 is not of type 'string'",
            ),
            (
                {"rh_subscription": {"add_pool": ["1"]}},
                None,
            ),
            # The json schema error message is not descriptive
            # but basically we need to confirm the schema will fail
            # the config validation when both add_pool and the deprecated
            # add-pool are added
            (
                {"rh_subscription": {"add_pool": ["1"], "add-pool": ["2"]}},
                r"({'add_pool': \['1'\], 'add-pool': \['2'\]} should not be"
                r" valid under {'required': \['add_pool', 'add-pool'\]}|"
                r"{'required': \['add_pool', 'add-pool'\]} is not allowed"
                r" for {'add_pool': \['1'\], 'add-pool': \['2'\]})",
            ),
            (
                {"rh_subscription": {"enable_repo": "name"}},
                "'name' is not of type 'array'",
            ),
            (
                {"rh_subscription": {"disable_repo": "name"}},
                "'name' is not of type 'array'",
            ),
            (
                {"rh_subscription": {"release_version": [10]}},
                r"\[10\] is not of type 'string'",
            ),
            (
                {
                    "rh_subscription": {
                        "activation_key": "foobar",
                        "org": "ABC",
                    }
                },
                None,
            ),
            (
                {"rh_subscription": {"activation_key": "foobar", "org": 314}},
                "Deprecated in version 24.2. Use of type integer for this"
                " value is deprecated. Use a string instead.",
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        self._validation_steps(config, error_msg)

    @pytest.mark.parametrize(
        "config, error_msg",
        [
            (
                {"rh_subscription": {"add-pool": ["1"]}},
                # The deprecation is not raised for jsonschema<4.0
                # as the latter can't merge $defs and inline keys
                r"Deprecated in version 25.3. Use \*\*add_pool\*\* instead.",
            ),
        ],
    )
    @skipUnlessJsonSchemaVersionGreaterThan(version=(3, 2, 0))
    def test_schema_validation_requiring_new_json_schema(
        self, config, error_msg
    ):
        self._validation_steps(config, error_msg)

    @staticmethod
    def _validation_steps(config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)


class TestConstructor:
    """
    Test Constructor operations
    """

    def test_deprecated_values(self):
        """
        Confirm the constructor assigns the deprecated fields' cfg keys to the
        correct python object fields
        """

        cfg_with_new_keys = {"rh_subscription": {}}
        cfg_with_deprecated_keys = {"rh_subscription": {}}

        deprecation_pairs = [
            ("activation-key", "activation_key"),
            ("disable-repo", "disable_repo"),
            ("enable-repo", "enable_repo"),
            ("add-pool", "add_pool"),
            ("rhsm-baseurl", "rhsm_baseurl"),
            ("server-hostname", "server_hostname"),
            ("auto-attach", "auto_attach"),
            ("service-level", "service_level"),
        ]

        counter = 0
        for tuple in deprecation_pairs:
            cfg_with_new_keys["rh_subscription"][tuple[0]] = counter
            cfg_with_deprecated_keys["rh_subscription"][tuple[1]] = counter
            counter = counter + 1

        mgr_with_new_keys = cc_rh_subscription.SubscriptionManager(
            cfg_with_new_keys
        )
        mgr_with_deprecated_keys = cc_rh_subscription.SubscriptionManager(
            cfg_with_deprecated_keys
        )

        assert (
            mgr_with_new_keys.rhel_cfg == cfg_with_new_keys["rh_subscription"]
        )
        assert (
            mgr_with_deprecated_keys.rhel_cfg
            == cfg_with_deprecated_keys["rh_subscription"]
        )

        dict_new_without_rhel_cfg = mgr_with_new_keys.__dict__
        del dict_new_without_rhel_cfg["rhel_cfg"]

        dict_deprecated_without_rhel_cfg = mgr_with_deprecated_keys.__dict__
        del dict_deprecated_without_rhel_cfg["rhel_cfg"]

        assert dict_new_without_rhel_cfg == dict_deprecated_without_rhel_cfg



# This file is part of cloud-init. See LICENSE file for license information.

import os
import textwrap

import pytest
import yaml

from cloudinit.sources.DataSourceNoCloud import DataSourceNoCloud as dsNoCloud
from cloudinit.sources.DataSourceNoCloud import (
    DataSourceNoCloudNet,
    parse_cmdline_data,
)
from tests.unittests.helpers import mock, populate_dir


@pytest.fixture(autouse=True)
def common_mocks(mocker):
    mocker.patch("cloudinit.sources.DataSourceNoCloud.util.is_lxd")
    mocker.patch("cloudinit.util.get_cmdline", return_value="root=TESTCMDLINE")
    mocker.patch("cloudinit.dmi.read_dmi_data", return_value=None)


class TestNoCloudDataSource:
    def _test_fs_config_is_read(
        self, fs_label, fs_label_to_search, mocker, paths
    ):
        vfat_device = "device-1"

        def m_mount_cb(device, callback, mtype):
            if device == vfat_device:
                return {"meta-data": yaml.dump({"instance-id": "IID"})}
            else:
                return {}

        def m_find_devs_with(query="", path=""):
            if "TYPE=vfat" == query:
                return [vfat_device]
            elif "LABEL={}".format(fs_label) == query:
                return [vfat_device]
            else:
                return []

        mocker.patch(
            "cloudinit.util.find_devs_with", side_effect=m_find_devs_with
        )
        mocker.patch("cloudinit.util.mount_cb", side_effect=m_mount_cb)
        sys_cfg = {"datasource": {"NoCloud": {"fs_label": fs_label_to_search}}}
        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()

        assert dsrc.metadata.get("instance-id") == "IID"
        assert ret

    def test_nocloud_seed_dir_on_lxd(self, paths):
        md = {"instance-id": "IID", "dsmode": "local"}
        ud = b"USER_DATA_HERE"
        seed_dir = os.path.join(paths.seed_dir, "nocloud")
        populate_dir(
            seed_dir, {"user-data": ud, "meta-data": yaml.safe_dump(md)}
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert dsrc.userdata_raw == ud
        assert dsrc.metadata == md
        assert dsrc.platform_type == "lxd"
        assert dsrc.subplatform == "seed-dir (%s)" % seed_dir
        assert ret

    def test_nocloud_seed_dir_non_lxd_platform_is_nocloud(self, mocker, paths):
        """Non-lxd environments will list nocloud as the platform."""
        mocker.patch(
            "cloudinit.sources.DataSourceNoCloud.util.is_lxd",
            return_value=False,
        )
        md = {"instance-id": "IID", "dsmode": "local"}
        seed_dir = os.path.join(paths.seed_dir, "nocloud")
        populate_dir(
            seed_dir, {"user-data": "", "meta-data": yaml.safe_dump(md)}
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        assert dsrc.get_data()
        assert dsrc.platform_type == "nocloud"
        assert dsrc.subplatform == "seed-dir (%s)" % seed_dir

    def test_nocloud_seedfrom(self, paths, caplog):
        """Check that a seedfrom triggers detection"""
        ds = DataSourceNoCloudNet(
            sys_cfg={"datasource": {"NoCloud": {"seedfrom": "somevalue"}}},
            distro=None,
            paths=paths,
        )
        assert ds.ds_detect()
        assert (
            "Machine is configured by system configuration to run on "
            "single datasource DataSourceNoCloudNet"
        ) in caplog.text

    def test_nocloud_user_data_meta_data(self, paths):
        """Check that meta-data and user-data trigger detection"""
        assert dsNoCloud(
            sys_cfg={
                "datasource": {
                    "NoCloud": {
                        "meta-data": "",
                        "user-data": "#cloud-config\nsome-config",
                    }
                }
            },
            distro=None,
            paths=paths,
        ).ds_detect()

    def test_fs_label(self, paths, mocker):
        # find_devs_with should not be called ff fs_label is None
        class PsuedoException(Exception):
            pass

        mocker.patch(
            "cloudinit.util.find_devs_with", side_effect=PsuedoException
        )

        # by default, NoCloud should search for filesystems by label
        sys_cfg = {"datasource": {"NoCloud": {}}}
        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        pytest.raises(PsuedoException, dsrc.get_data)

        # but disabling searching should just end up with None found
        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}
        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert not ret

    def test_fs_config_lowercase_label(self, mocker, paths):
        self._test_fs_config_is_read("cidata", "cidata", mocker, paths)

    def test_fs_config_uppercase_label(self, mocker, paths):
        self._test_fs_config_is_read("CIDATA", "cidata", mocker, paths)

    def test_fs_config_lowercase_label_search_uppercase(self, mocker, paths):
        self._test_fs_config_is_read("cidata", "CIDATA", mocker, paths)

    def test_fs_config_uppercase_label_search_uppercase(self, mocker, paths):
        self._test_fs_config_is_read("CIDATA", "CIDATA", mocker, paths)

    def test_no_datasource_expected(self, paths):
        # no source should be found if no cmdline, config, and fs_label=None
        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        assert not dsrc.get_data()

    def test_seed_in_config(self, paths):
        data = {
            "fs_label": None,
            "meta-data": yaml.safe_dump({"instance-id": "IID"}),
            "user-data": b"USER_DATA_RAW",
        }

        sys_cfg = {"datasource": {"NoCloud": data}}
        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert dsrc.userdata_raw == b"USER_DATA_RAW"
        assert dsrc.metadata.get("instance-id") == "IID"
        assert ret

    def test_nocloud_seed_with_vendordata(self, paths):
        md = {"instance-id": "IID", "dsmode": "local"}
        ud = b"USER_DATA_HERE"
        vd = b"THIS IS MY VENDOR_DATA"

        populate_dir(
            os.path.join(paths.seed_dir, "nocloud"),
            {
                "user-data": ud,
                "meta-data": yaml.safe_dump(md),
                "vendor-data": vd,
            },
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert dsrc.userdata_raw == ud
        assert dsrc.metadata == md
        assert dsrc.vendordata_raw == vd
        assert ret

    def test_nocloud_no_vendordata(self, paths):
        populate_dir(
            os.path.join(paths.seed_dir, "nocloud"),
            {"user-data": b"ud", "meta-data": "instance-id: IID\n"},
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert dsrc.userdata_raw == b"ud"
        assert not dsrc.vendordata
        assert ret

    def test_metadata_network_interfaces(self, paths):
        gateway = "103.225.10.1"
        md = {
            "instance-id": "i-abcd",
            "local-hostname": "hostname1",
            "network-interfaces": textwrap.dedent(
                """\
                auto eth0
                iface eth0 inet static
                hwaddr 00:16:3e:70:e1:04
                address 103.225.10.12
                netmask 255.255.255.0
                gateway """
                + gateway
                + """
                dns-servers 8.8.8.8"""
            ),
        }

        populate_dir(
            os.path.join(paths.seed_dir, "nocloud"),
            {"user-data": b"ud", "meta-data": yaml.dump(md) + "\n"},
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert ret
        # very simple check just for the strings above
        assert gateway in str(dsrc.network_config)

    def test_metadata_network_config(self, paths):
        # network-config needs to get into network_config
        netconf = {
            "version": 1,
            "config": [
                {
                    "type": "physical",
                    "name": "interface0",
                    "subnets": [{"type": "dhcp"}],
                }
            ],
        }
        populate_dir(
            os.path.join(paths.seed_dir, "nocloud"),
            {
                "user-data": b"ud",
                "meta-data": "instance-id: IID\n",
                "network-config": yaml.dump(netconf) + "\n",
            },
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert ret
        assert netconf == dsrc.network_config

    def test_metadata_network_config_over_interfaces(self, paths):
        # network-config should override meta-data/network-interfaces
        gateway = "103.225.10.1"
        md = {
            "instance-id": "i-abcd",
            "local-hostname": "hostname1",
            "network-interfaces": textwrap.dedent(
                """\
                auto eth0
                iface eth0 inet static
                hwaddr 00:16:3e:70:e1:04
                address 103.225.10.12
                netmask 255.255.255.0
                gateway """
                + gateway
                + """
                dns-servers 8.8.8.8"""
            ),
        }

        netconf = {
            "version": 1,
            "config": [
                {
                    "type": "physical",
                    "name": "interface0",
                    "subnets": [{"type": "dhcp"}],
                }
            ],
        }
        populate_dir(
            os.path.join(paths.seed_dir, "nocloud"),
            {
                "user-data": b"ud",
                "meta-data": yaml.dump(md) + "\n",
                "network-config": yaml.dump(netconf) + "\n",
            },
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc.get_data()
        assert ret
        assert netconf == dsrc.network_config
        assert gateway not in str(dsrc.network_config)

    @mock.patch("cloudinit.util.blkid")
    def test_nocloud_get_devices_freebsd(self, fake_blkid, mocker, paths):
        populate_dir(
            os.path.join(paths.seed_dir, "nocloud"),
            {"user-data": b"ud", "meta-data": "instance-id: IID\n"},
        )

        sys_cfg = {"datasource": {"NoCloud": {"fs_label": None}}}

        mocker.patch("cloudinit.util.is_FreeBSD", return_value=True)

        def _mfind_devs_with_freebsd(
            criteria=None,
            oformat="device",
            tag=None,
            no_cache=False,
            path=None,
        ):
            if not criteria:
                return ["/dev/msdosfs/foo", "/dev/iso9660/foo"]
            if criteria.startswith("LABEL="):
                return ["/dev/msdosfs/foo", "/dev/iso9660/foo"]
            elif criteria == "TYPE=vfat":
                return ["/dev/msdosfs/foo"]
            elif criteria == "TYPE=iso9660":
                return ["/dev/iso9660/foo"]
            return []

        mocker.patch(
            "cloudinit.util.find_devs_with",
            side_effect=_mfind_devs_with_freebsd,
        )

        dsrc = dsNoCloud(sys_cfg=sys_cfg, distro=None, paths=paths)
        ret = dsrc._get_devices("foo")
        assert ["/dev/msdosfs/foo", "/dev/iso9660/foo"] == ret
        fake_blkid.assert_not_called()


class TestParseCommandLineData:
    def test_parse_cmdline_data_valid(self):
        ds_id = "ds=nocloud"
        pairs = (
            ("root=/dev/sda1 %(ds_id)s", {}),
            ("%(ds_id)s; root=/dev/foo", {}),
            ("%(ds_id)s", {}),
            ("%(ds_id)s;", {}),
            ("%(ds_id)s;s=SEED", {"seedfrom": "SEED"}),
            (
                "%(ds_id)s;seedfrom=SEED;local-hostname=xhost",
                {"seedfrom": "SEED", "local-hostname": "xhost"},
            ),
            ("%(ds_id)s;h=xhost", {"local-hostname": "xhost"}),
            (
                "%(ds_id)s;h=xhost;i=IID",
                {"local-hostname": "xhost", "instance-id": "IID"},
            ),
        )

        for fmt, expected in pairs:
            fill = {}
            cmdline = fmt % {"ds_id": ds_id}
            ret = parse_cmdline_data(ds_id=ds_id, fill=fill, cmdline=cmdline)
            assert expected == fill
            assert ret

    def test_parse_cmdline_data_none(self):
        ds_id = "ds=foo"
        cmdlines = (
            "root=/dev/sda1 ro",
            "console=/dev/ttyS0 root=/dev/foo",
            "",
            "ds=foocloud",
            "ds=foo-net",
            "ds=nocloud;s=SEED",
        )

        for cmdline in cmdlines:
            fill = {}
            ret = parse_cmdline_data(ds_id=ds_id, fill=fill, cmdline=cmdline)
            assert fill == {}
            assert not ret



"""ansible enables running on first boot either ansible-pull"""

import abc
import logging
import os
import re
import sys
import sysconfig
from copy import deepcopy
from typing import List, Optional

from cloudinit import lifecycle, subp
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema
from cloudinit.distros import ALL_DISTROS, Distro
from cloudinit.settings import PER_INSTANCE
from cloudinit.util import get_cfg_by_path

meta: MetaSchema = {
    "id": "cc_ansible",
    "frequency": PER_INSTANCE,
    "distros": [ALL_DISTROS],
    "activate_by_schema_keys": ["ansible"],
}

LOG = logging.getLogger(__name__)
CFG_OVERRIDE = "ansible_config"


class AnsiblePull(abc.ABC):
    def __init__(self, distro: Distro):
        self.cmd_pull = ["ansible-pull"]
        self.cmd_version = ["ansible-pull", "--version"]
        self.distro = distro
        self.env = {}
        self.run_user: Optional[str] = None

        # some ansible modules directly reference os.environ["HOME"]
        # and cloud-init might not have that set, default: /root
        self.env["HOME"] = os.environ.get("HOME", "/root")

    def get_version(self) -> Optional[lifecycle.Version]:
        stdout, _ = self.do_as(self.cmd_version)
        first_line = stdout.splitlines().pop(0)
        matches = re.search(r"([\d\.]+)", first_line)
        if matches:
            version = matches.group(0)
            return lifecycle.Version.from_str(version)
        return None

    def pull(self, *args) -> str:
        stdout, _ = self.do_as([*self.cmd_pull, *args])
        return stdout

    def check_deps(self):
        if not self.is_installed():
            raise ValueError("command: ansible is not installed")

    def do_as(self, command: list, **kwargs):
        if not self.run_user:
            return self.subp(command, **kwargs)
        return self.distro.do_as(command, self.run_user, **kwargs)

    def subp(self, command, **kwargs):
        return subp.subp(command, update_env=self.env, **kwargs)

    @abc.abstractmethod
    def is_installed(self):
        pass

    @abc.abstractmethod
    def install(self, pkg_name: str):
        pass


class AnsiblePullPip(AnsiblePull):
    def __init__(self, distro: Distro, user: Optional[str]):
        super().__init__(distro)
        self.run_user = user
        self.add_pip_install_site_to_path()

    def add_pip_install_site_to_path(self):
        if self.run_user:
            user_base, _ = self.do_as(
                [
                    sys.executable,
                    "-c",
                    "import site; print(site.getuserbase())",
                ]
            )
            ansible_path = f"{user_base}/bin/"

            old_path = self.env.get("PATH")
            if old_path:
                self.env["PATH"] = ":".join([old_path, ansible_path])
            else:
                self.env["PATH"] = ansible_path

    def bootstrap_pip_if_required(self):
        try:
            import pip  # noqa: F401
        except ImportError:
            self.distro.install_packages([self.distro.pip_package_name])

    def install(self, pkg_name: str):
        """should cloud-init grow an interface for non-distro package
        managers? this seems reusable
        """
        self.bootstrap_pip_if_required()

        if not self.is_installed():
            cmd = [
                sys.executable,
                "-m",
                "pip",
                "install",
            ]

            if os.path.exists(
                os.path.join(
                    sysconfig.get_path("stdlib"), "EXTERNALLY-MANAGED"
                )
            ):
                cmd.append("--break-system-packages")
            if self.run_user:
                cmd.append("--user")

            self.__upgrade_pip(cmd)
            LOG.info("Installing the %s package", pkg_name)
            self.do_as([*cmd, pkg_name])
            LOG.info("Installed the %s package", pkg_name)

    def __upgrade_pip(self, cmd: list):
        """Try to upgrade pip then not raise an exception
        if the pip upgrade fails"""
        LOG.info("Upgrading pip")
        try:
            self.do_as([*cmd, "--upgrade", "pip"])
        except subp.ProcessExecutionError as e:
            LOG.warning(
                (
                    "Failed at upgrading pip. This is usually not critical"
                    "so the script will skip this step.\n%s"
                ),
                e,
            )
        else:
            LOG.info("Upgraded pip")

    def is_installed(self) -> bool:
        cmd = [sys.executable, "-m", "pip", "list"]

        if self.run_user:
            cmd.append("--user")

        stdout, _ = self.do_as(cmd)
        return "ansible" in stdout


class AnsiblePullDistro(AnsiblePull):
    def __init__(self, distro: Distro, user: Optional[str]):
        super().__init__(distro)
        self.run_user = user

    def install(self, pkg_name: str):
        if not self.is_installed():
            self.distro.install_packages([pkg_name])

    def is_installed(self) -> bool:
        return bool(subp.which("ansible"))


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:

    ansible_cfg: dict = cfg.get("ansible", {})
    ansible_user = ansible_cfg.get("run_user")
    install_method = ansible_cfg.get("install_method")
    setup_controller = ansible_cfg.get("setup_controller")

    galaxy_cfg = ansible_cfg.get("galaxy")
    pull_cfg = ansible_cfg.get("pull")
    package_name = ansible_cfg.get("package_name", "")

    if ansible_cfg:
        ansible: AnsiblePull
        validate_config(ansible_cfg)

        distro: Distro = cloud.distro
        if install_method == "pip":
            ansible = AnsiblePullPip(distro, ansible_user)
        else:
            ansible = AnsiblePullDistro(distro, ansible_user)
        ansible.install(package_name)
        ansible.check_deps()
        ansible_config = ansible_cfg.get("ansible_config", "")

        if ansible_config:
            ansible.env[CFG_OVERRIDE] = ansible_config

        if galaxy_cfg:
            ansible_galaxy(galaxy_cfg, ansible)

        if pull_cfg:
            if isinstance(pull_cfg, dict):
                pull_cfg = [pull_cfg]
            for cfg in pull_cfg:
                run_ansible_pull(ansible, deepcopy(cfg))

        if setup_controller:
            ansible_controller(setup_controller, ansible)


def validate_config(cfg: dict):
    required_keys = (
        "install_method",
        "package_name",
    )
    for key in required_keys:
        if not get_cfg_by_path(cfg, key):
            raise ValueError(f"Missing required key '{key}' from {cfg}")
    pull_cfg = cfg.get("pull")
    if pull_cfg:
        if isinstance(pull_cfg, dict):
            pull_cfg = [pull_cfg]
        elif not isinstance(pull_cfg, list):
            raise ValueError(
                "Invalid value ansible.pull. Expected either dict of list of"
                f" dicts but found {pull_cfg}"
            )
        for p_cfg in pull_cfg:
            if not isinstance(p_cfg, dict):
                raise ValueError(
                    "Invalid value of ansible.pull. Expected dict but found"
                    f" {p_cfg}"
                )
            if not get_cfg_by_path(p_cfg, "url"):
                raise ValueError(f"Missing required key 'url' from {p_cfg}")
            has_playbook = get_cfg_by_path(p_cfg, "playbook_name")
            has_playbooks = get_cfg_by_path(p_cfg, "playbook_names")
            if not any([has_playbook, has_playbooks]):
                raise ValueError(
                    f"Missing required key 'playbook_names' from {p_cfg}"
                )
            elif all([has_playbooks, has_playbook]):
                raise ValueError(
                    "Key 'ansible.pull.playbook_name' and"
                    " 'ansible.pull.playbook_names' are mutually exclusive."
                    f" Please use 'playbook_names' in {p_cfg}"
                )

    controller_cfg = cfg.get("setup_controller")
    if controller_cfg:
        if not any(
            [
                controller_cfg.get("repositories"),
                controller_cfg.get("run_ansible"),
            ]
        ):
            raise ValueError(f"Missing required key from {controller_cfg}")

    install = cfg["install_method"]
    if install not in ("pip", "distro"):
        raise ValueError("Invalid install method {install}")


def filter_args(cfg: dict) -> dict:
    """remove boolean false values"""
    return {
        key.replace("_", "-"): value
        for (key, value) in cfg.items()
        if value is not False
    }


def run_ansible_pull(pull: AnsiblePull, cfg: dict):
    playbook_name: str = cfg.pop("playbook_name", None)
    playbook_names: List[str] = cfg.pop("playbook_names", None)
    if playbook_name:
        playbook_names = [playbook_name]
    v = pull.get_version()
    if not v:
        LOG.warning("Cannot parse ansible version")
    elif v < lifecycle.Version(2, 7, 0):
        # diff was added in commit edaa0b52450ade9b86b5f63097ce18ebb147f46f
        if cfg.get("diff"):
            raise ValueError(
                f"Ansible version {v.major}.{v.minor}.{v.patch}"
                "doesn't support --diff flag, exiting."
            )
    pull_args = [
        f"--{key}={value}" if value is not True else f"--{key}"
        for key, value in filter_args(cfg).items()
    ]
    if v and v >= lifecycle.Version(2, 12, 0):
        # Multiple playbook support was resolved in 2.12.
        # https://github.com/ansible/ansible/pull/73172
        stdout = pull.pull(*pull_args, *playbook_names)
        if stdout:
            sys.stdout.write(f"{stdout}")
    else:
        # Older ansible must pull separate playbooks
        for playbook in playbook_names:
            stdout = pull.pull(*pull_args, playbook)
            if stdout:
                sys.stdout.write(f"{stdout}")


def ansible_galaxy(cfg: dict, ansible: AnsiblePull):
    actions = cfg.get("actions", [])

    if not actions:
        LOG.warning("Invalid config: %s", cfg)
    for command in actions:
        ansible.do_as(command)


def ansible_controller(cfg: dict, ansible: AnsiblePull):
    for repository in cfg.get("repositories", []):
        ansible.do_as(
            ["git", "clone", repository["source"], repository["path"]]
        )
    for args in cfg.get("run_ansible", []):
        playbook_dir = args.pop("playbook_dir")
        playbook_name = args.pop("playbook_name")
        command = [
            "ansible-playbook",
            playbook_name,
            *[f"--{key}={value}" for key, value in filter_args(args).items()],
        ]
        ansible.do_as(command, cwd=playbook_dir)



# This file is part of cloud-init. See LICENSE file for license information.


import copy
from unittest import mock

import pytest

from cloudinit import distros, importer, sources
from cloudinit.sources import DataSourceCloudSigma
from cloudinit.sources.helpers.cloudsigma import Cepko

SERVER_CONTEXT = {
    "cpu": 1000,
    "cpus_instead_of_cores": False,
    "mem": 1073741824,
    "meta": {
        "ssh_public_key": "ssh-rsa AAAAB3NzaC1yc2E.../hQ5D5 john@doe",
        "cloudinit-user-data": "#cloud-config\n\n...",
    },
    "name": "test_server",
    "requirements": [],
    "smp": 1,
    "tags": ["much server", "very performance"],
    "uuid": "65b2fb23-8c03-4187-a3ba-8b7c919e8890",
    "vnc_password": "9e84d6cb49e46379",
    "vendor_data": {
        "location": "zrh",
        "cloudinit": "#cloud-config\n\n...",
    },
}

DS_PATH = "cloudinit.sources.DataSourceCloudSigma.DataSourceCloudSigma"


class CepkoMock(Cepko):
    def __init__(self, mocked_context):
        self.result = mocked_context

    def all(self):
        return self


@pytest.fixture
def ds(mocker, paths):
    mocker.patch(DS_PATH + ".override_ds_detect", return_value=True)
    distro_cls = distros.fetch("ubuntu")
    distro = distro_cls("ubuntu", cfg={}, paths=paths)
    datasource = DataSourceCloudSigma.DataSourceCloudSigma(
        sys_cfg={}, distro=distro, paths=paths
    )
    datasource.cepko = CepkoMock(SERVER_CONTEXT)
    return datasource


class TestDataSourceCloudSigma:

    def test_get_hostname(self, ds):
        ds.get_data()
        assert "test_server" == ds.get_hostname().hostname
        ds.metadata["name"] = ""
        assert "65b2fb23" == ds.get_hostname().hostname
        utf8_hostname = b"\xd1\x82\xd0\xb5\xd1\x81\xd1\x82".decode("utf-8")
        ds.metadata["name"] = utf8_hostname
        assert "65b2fb23" == ds.get_hostname().hostname

    def test_get_public_ssh_keys(self, ds):
        ds.get_data()
        assert [
            SERVER_CONTEXT["meta"]["ssh_public_key"]
        ] == ds.get_public_ssh_keys()

    def test_get_instance_id(self, ds):
        ds.get_data()
        assert SERVER_CONTEXT["uuid"] == ds.get_instance_id()

    def test_platform(self, ds):
        """All platform-related attributes are set."""
        ds.get_data()
        assert ds.cloud_name == "cloudsigma"
        assert ds.platform_type == "cloudsigma"
        assert ds.subplatform == "cepko (/dev/ttyS1)"

    def test_metadata(self, ds):
        ds.get_data()
        assert ds.metadata == SERVER_CONTEXT

    def test_user_data(self, ds):
        ds.get_data()
        assert ds.userdata_raw == SERVER_CONTEXT["meta"]["cloudinit-user-data"]

    def test_encoded_user_data(self, ds):
        encoded_context = copy.deepcopy(SERVER_CONTEXT)
        encoded_context["meta"]["base64_fields"] = "cloudinit-user-data"
        encoded_context["meta"]["cloudinit-user-data"] = "aGkgd29ybGQK"
        ds.cepko = CepkoMock(encoded_context)
        ds.get_data()

        assert ds.userdata_raw == b"hi world\n"

    def test_vendor_data(self, ds):
        ds.get_data()
        assert ds.vendordata_raw == SERVER_CONTEXT["vendor_data"]["cloudinit"]

    def test_lack_of_vendor_data(self, ds):
        stripped_context = copy.deepcopy(SERVER_CONTEXT)
        del stripped_context["vendor_data"]
        ds.cepko = CepkoMock(stripped_context)
        ds.get_data()

        assert ds.vendordata_raw is None

    def test_lack_of_cloudinit_key_in_vendor_data(self, ds):
        stripped_context = copy.deepcopy(SERVER_CONTEXT)
        del stripped_context["vendor_data"]["cloudinit"]
        ds.cepko = CepkoMock(stripped_context)
        ds.get_data()

        assert ds.vendordata_raw is None


class TestDsLoads:
    def test_get_datasource_list_returns_in_local(self):
        deps = (sources.DEP_FILESYSTEM,)
        ds_list = DataSourceCloudSigma.get_datasource_list(deps)
        assert ds_list == [DataSourceCloudSigma.DataSourceCloudSigma]

    @mock.patch.object(
        importer,
        "match_case_insensitive_module_name",
        lambda name: f"DataSource{name}",
    )
    def test_list_sources_finds_ds(self):
        found = sources.list_sources(
            ["CloudSigma"],
            (sources.DEP_FILESYSTEM,),
            ["cloudinit.sources"],
        )
        assert [DataSourceCloudSigma.DataSourceCloudSigma] == found
